"""
Directed Computing Graph package for sera.

This package provides classes and utilities for working with directed computing graphs.
"""

# Import type aliases and annotated types
# Import all classes from submodules
from ._dcg import DirectedComputingGraph, NodeId, TaskArgs, TaskKey
from ._edge import DCGEdge
from ._flow import Flow
from ._node import ComputeFn, ComputeFnId, DCGNode, PartialFn
from ._runtime import SKIP, UNSET, ArgValueType, NodeRuntime

# Import utility functions from type conversion
from ._type_conversion import (
    ComposeTypeConversion,
    TypeConversion,
    UnitTypeConversion,
    align_generic_type,
    ground_generic_type,
    is_generic_type,
    patch_get_origin,
)

# Define __all__ to control what gets exported
__all__ = [
    # Main classes
    "DirectedComputingGraph",
    "DCGNode",
    "DCGEdge",
    "Flow",
    "PartialFn",
    "TypeConversion",
    "NodeRuntime",
    # Enums and special values
    "ArgValueType",
    "UNSET",
    "SKIP",
    # Type aliases and annotations
    "NodeId",
    "TaskKey",
    "TaskArgs",
    "ComputeFnId",
    "ComputeFn",
    "UnitTypeConversion",
    "ComposeTypeConversion",
    # Utility functions
    "patch_get_origin",
    "is_generic_type",
    "align_generic_type",
    "ground_generic_type",
]



from __future__ import annotations

from typing import Any, Callable, Optional

from sera.libs.directed_computing_graph._node import ComputeFn, ComputeFnId


class Flow:
    def __init__(
        self,
        source: list[ComputeFnId] | ComputeFnId,
        target: ComputeFn,
        filter_fn: Optional[Callable[[Any], bool]] = None,
    ):
        self.source = [source] if isinstance(source, str) else source
        self.target = target
        self.filter_fn = filter_fn



from __future__ import annotations

from collections import Counter
from dataclasses import dataclass

from sera.misc import to_snake_case
from sera.models._multi_lingual_string import MultiLingualString


@dataclass
class EnumValue:
    name: str
    value: str | int
    description: MultiLingualString


@dataclass
class Enum:
    """Enum class to represent a set of named values."""

    # name of the enum in the application layer
    name: str
    values: dict[str, EnumValue]

    def __post_init__(self):
        # Ensure that all `value` attributes in EnumValue are unique
        unique_values = {value.value for value in self.values.values()}
        if len(unique_values) != len(self.values):
            value_counts = Counter(value.value for value in self.values.values())
            duplicates = [val for val, count in value_counts.items() if count > 1]
            raise ValueError(
                f"All 'value' attributes in EnumValue must be unique. Duplicates found: {duplicates}"
            )

        # Ensure that all `value` attributes in EnumValue are either all strings or all integers
        if not (self.is_str_enum() or self.is_int_enum()):
            raise ValueError(
                "All 'value' attributes in EnumValue must be either all strings or all integers."
            )

    def get_pymodule_name(self) -> str:
        """Get the python module name of this enum as if there is a python module created to store this enum only."""
        return to_snake_case(self.name)

    def get_tsmodule_name(self) -> str:
        """Get the typescript module name of this enum as if there is a typescript module created to store this enum only."""
        return self.name[0].lower() + self.name[1:]

    def is_str_enum(self) -> bool:
        """Check if this enum is a string enum."""
        return all(isinstance(value.value, str) for value in self.values.values())

    def is_int_enum(self) -> bool:
        """Check if this enum is a int enum."""
        return all(isinstance(value.value, int) for value in self.values.values())



from __future__ import annotations

import collections.abc
import inspect
from locale import normalize
from types import UnionType
from typing import (
    Annotated,
    Any,
    Callable,
    Mapping,
    MutableMapping,
    MutableSequence,
    MutableSet,
    Sequence,
    Set,
    TypeVar,
    Union,
    cast,
    get_args,
    get_origin,
    get_type_hints,
)

from sera.misc import identity

UnitTypeConversion = Annotated[
    Callable[[Any], Any], "A function that convert an object of type T1 to T2"
]
ComposeTypeConversion = Annotated[
    Callable[[Any, UnitTypeConversion], Any],
    "A function that convert a generic object of type G[T1] to G[T2]",
]


class TypeConversion:
    """Inspired by Rust type conversion traits. This class allows to derive a type conversion function from output of a pipe to input of another pipe."""

    class UnknownConversion(Exception):
        pass

    def __init__(
        self, type_casts: Sequence[UnitTypeConversion | ComposeTypeConversion]
    ):
        self.generic_single_type_conversion: dict[type, UnitTypeConversion] = {}
        self.unit_type_conversions: dict[tuple[type, type], UnitTypeConversion] = {}
        self.compose_type_conversion: dict[type, ComposeTypeConversion] = {}

        for fn in type_casts:
            assert not inspect.iscoroutinefunction(
                fn
            ), "Async conversion functions are not supported"
            sig = get_type_hints(fn)
            if len(sig) == 2:
                fn = cast(UnitTypeConversion, fn)

                intype = sig[[x for x in sig if x != "return"][0]]
                outtype = sig["return"]

                intype_origin = get_origin(intype)
                intype_args = get_args(intype)
                if (
                    intype_origin is not None
                    and len(intype_args) == 1
                    and intype_args[0] is outtype
                    and isinstance(outtype, TypeVar)
                ):
                    # this is a generic conversion G[T] => T
                    self.generic_single_type_conversion[intype_origin] = fn
                else:
                    self.unit_type_conversions[intype, outtype] = fn
            else:
                assert len(sig) == 3, "Invalid type conversion function"
                fn = cast(ComposeTypeConversion, fn)

                intype = sig[[x for x in sig if x != "return"][0]]
                outtype = sig["return"]
                intype_origin = get_origin(intype)
                assert intype_origin is not None
                self.compose_type_conversion[intype_origin] = fn

    def get_conversion(
        self, source_type: type, target_type: type
    ) -> UnitTypeConversion:
        # handle identity conversion
        # happen when source_type = target_type or target_type is Union[source_type, ...]
        if source_type == target_type:
            # source_type is target_type doesn't work with collections.abc.Sequence
            return identity
        if get_origin(target_type) in (Union, UnionType) and source_type in get_args(
            target_type
        ):
            return identity

        if (source_type, target_type) in self.unit_type_conversions:
            # we already have a unit type conversion function for these types
            return self.unit_type_conversions[source_type, target_type]

        # check if this is a generic conversion
        intype_origin = get_origin(source_type)
        intype_args = get_args(source_type)

        if intype_origin is None or len(intype_args) != 1:
            raise TypeConversion.UnknownConversion(
                f"Cannot find conversion from {source_type} to {target_type}"
            )

        outtype_origin = get_origin(target_type)
        outtype_args = get_args(target_type)

        if outtype_origin is None:
            # we are converting G[T] => T'
            if (
                target_type is not intype_args[0]
                or intype_origin not in self.generic_single_type_conversion
            ):
                # either T != T' or G is unkknown
                raise TypeConversion.UnknownConversion(
                    f"Cannot find conversion from {source_type} to {target_type}"
                )
            return self.generic_single_type_conversion[intype_origin]

        # we are converting G[T] => G'[T']
        if (
            outtype_origin is not intype_origin
            or intype_origin not in self.compose_type_conversion
        ):
            # either G != G' or G is unknown
            raise TypeConversion.UnknownConversion(
                f"Cannot find conversion from {source_type} to {target_type}"
            )
        # G == G' => T == T'
        compose_func = self.compose_type_conversion[intype_origin]
        func = self.get_conversion(intype_args[0], outtype_args[0])
        return lambda x: compose_func(x, func)


def patch_get_origin(t: type) -> Any:
    """The original get_origin(typing.Sequence) returns collections.abc.Sequence.
    Later comparing typing.Sequence[T] to collections.abc.Sequence[T] aren't equal.

    This function will return typing.Sequence instead.
    """
    origin = get_origin(t)
    if origin is None:
        return origin
    return {
        collections.abc.Mapping: Mapping,
        collections.abc.Sequence: Sequence,
        collections.abc.MutableSequence: MutableSequence,
        collections.abc.MutableMapping: MutableMapping,
        collections.abc.Set: Set,
        collections.abc.MutableSet: MutableSet,
    }.get(origin, origin)


def is_generic_type(t: type) -> bool:
    return isinstance(t, TypeVar) or any(is_generic_type(a) for a in get_args(t))


def align_generic_type(
    generic_type: type, target_type: type
) -> tuple[type, tuple[type, type]]:
    """Return the grounded outer type, and the mapping from the TypeVar to the concrete type"""
    if isinstance(generic_type, TypeVar):
        return target_type, (generic_type, target_type)

    origin = patch_get_origin(generic_type)
    assert origin is not None
    if origin != patch_get_origin(target_type):
        raise TypeConversion.UnknownConversion(
            f"Cannot ground generic type {generic_type} to {target_type}"
        )

    if len(get_args(generic_type)) != 1:
        raise NotImplementedError()

    gt = align_generic_type(get_args(generic_type)[0], get_args(target_type)[0])
    return origin[gt[0]], gt[1]


def ground_generic_type(generic_type: type, var2type: dict[TypeVar, type]) -> type:
    if isinstance(generic_type, TypeVar):
        return var2type[generic_type]

    origin = get_origin(generic_type)
    if origin is None:
        # nothing to ground
        return generic_type

    return origin[*(ground_generic_type(t, var2type) for t in get_args(generic_type))]



from __future__ import annotations

import re
from typing import Collection, Generic, cast

from litestar import Request, status_codes
from litestar.connection import ASGIConnection
from litestar.dto import MsgspecDTO
from litestar.dto._backend import DTOBackend
from litestar.dto._codegen_backend import DTOCodegenBackend
from litestar.enums import RequestEncodingType
from litestar.exceptions import HTTPException
from litestar.serialization import decode_json, decode_msgpack
from litestar.typing import FieldDefinition
from msgspec import Struct

from sera.libs.base_service import Query, QueryOp
from sera.libs.middlewares.uscp import STATE_SYSTEM_CONTROLLED_PROP_KEY
from sera.typing import T

# for parsing field names and operations from query string
FIELD_REG = re.compile(r"(?P<name>[a-zA-Z_0-9]+)(?:\[(?P<op>[a-zA-Z0-9]+)\])?")
QUERY_OPS = {op.value for op in QueryOp}
KEYWORDS = {"field", "limit", "offset", "unique", "sorted_by", "group_by"}


def parse_query(request: Request, fields: set[str], debug: bool) -> Query:
    """Parse query for retrieving records that match a query.

    If a field name collides with a keyword, you can add `_` to the field name.

    To filter records, you can apply a condition on a column using <field>=<value> (equal condition). Or you can
    be explicit by using <field>[op]=<value>, where op is one of the operators defined in QueryOp.
    """
    query: Query = {}

    for k, v in request.query_params.items():
        if k in KEYWORDS:
            continue
        m = FIELD_REG.match(k)
        if m:
            field_name = m.group("name")
            operation = m.group("op")  # This will be None if no operation is specified

            # If field name ends with '_' and it's to avoid keyword conflict, remove it
            if field_name.endswith("_") and field_name[:-1] in KEYWORDS:
                field_name = field_name[:-1]

            if field_name not in fields:
                # Invalid field name, skip
                if debug:
                    raise HTTPException(
                        status_code=status_codes.HTTP_400_BAD_REQUEST,
                        detail=f"Invalid field name: {field_name}",
                    )
                continue

            # Process based on operation or default to equality check
            if not operation:
                operation = QueryOp.eq
            else:
                if operation not in QUERY_OPS:
                    raise HTTPException(
                        status_code=status_codes.HTTP_400_BAD_REQUEST,
                        detail=f"Invalid operation: {operation}",
                    )
                operation = QueryOp(operation)
            query[field_name] = {operation: v}
        else:
            # Invalid field name format
            if debug:
                raise HTTPException(
                    status_code=status_codes.HTTP_400_BAD_REQUEST,
                    detail=f"Invalid field name: {k}",
                )
            continue

    return query


class SingleAutoUSCP(MsgspecDTO[T], Generic[T]):
    """Auto Update System Controlled Property DTO"""

    @classmethod
    def create_for_field_definition(
        cls,
        field_definition: FieldDefinition,
        handler_id: str,
        backend_cls: type[DTOBackend] | None = None,
    ) -> None:
        assert backend_cls is None, "Custom backend not supported"
        super().create_for_field_definition(
            field_definition, handler_id, FixedDTOBackend
        )

    def decode_bytes(self, value: bytes):
        """Decode a byte string into an object"""
        backend = self._dto_backends[self.asgi_connection.route_handler.handler_id][
            "data_backend"
        ]  # pyright: ignore
        obj = backend.populate_data_from_raw(value, self.asgi_connection)
        obj.update_system_controlled_props(
            self.asgi_connection.scope["state"][STATE_SYSTEM_CONTROLLED_PROP_KEY]
        )
        return obj


class FixedDTOBackend(DTOCodegenBackend):
    def parse_raw(
        self, raw: bytes, asgi_connection: ASGIConnection
    ) -> Struct | Collection[Struct]:
        """Parse raw bytes into transfer model type.

        Note: instead of decoding into self.annotation, which I encounter this error: https://github.com/litestar-org/litestar/issues/4181; we have to use self.model_type, which is the original type.

        Args:
            raw: bytes
            asgi_connection: The current ASGI Connection

        Returns:
            The raw bytes parsed into transfer model type.
        """
        request_encoding = RequestEncodingType.JSON

        if (content_type := getattr(asgi_connection, "content_type", None)) and (
            media_type := content_type[0]
        ):
            request_encoding = media_type

        type_decoders = asgi_connection.route_handler.resolve_type_decoders()

        if request_encoding == RequestEncodingType.MESSAGEPACK:
            result = decode_msgpack(
                value=raw,
                target_type=self.model_type,
                type_decoders=type_decoders,
                strict=False,
            )
        else:
            result = decode_json(
                value=raw,
                target_type=self.model_type,
                type_decoders=type_decoders,
                strict=False,
            )

        return cast("Struct | Collection[Struct]", result)



from sera.models._class import Class
from sera.models._collection import DataCollection
from sera.models._datatype import DataType, PyTypeWithDep, TsTypeWithDep
from sera.models._enum import Enum
from sera.models._module import App, Module, Package
from sera.models._multi_lingual_string import MultiLingualString
from sera.models._parse import parse_schema
from sera.models._property import Cardinality, DataProperty, ObjectProperty, Property
from sera.models._schema import Schema

__all__ = [
    "parse_schema",
    "Schema",
    "Property",
    "DataProperty",
    "ObjectProperty",
    "Class",
    "Cardinality",
    "DataType",
    "MultiLingualString",
    "Package",
    "DataCollection",
    "Module",
    "App",
    "PyTypeWithDep",
    "TsTypeWithDep",
    "Enum",
]



from __future__ import annotations

from litestar import Litestar
from litestar.testing import TestClient


def test_has(
    client: TestClient[Litestar],
    base_url: str,
    exist_records: list[int | str],
    non_exist_records: list[int | str],
) -> None:
    for record in exist_records:
        resp = client.head(f"{base_url}/{record}")
        assert (
            resp.status_code == 204
        ), f"Record {record} should exist but got {resp.status_code}"

    for record in non_exist_records:
        resp = client.head(f"{base_url}/{record}")
        assert (
            resp.status_code == 404
        ), f"Record {record} should not exist but got {resp.status_code}"


def test_get_by_id(
    client: TestClient[Litestar],
    base_url: str,
    exist_records: dict[int | str, dict],
    non_exist_records: list[int | str],
) -> None:
    for record, data in exist_records.items():
        resp = client.get(f"{base_url}/{record}")
        assert (
            resp.status_code == 200
        ), f"Record {record} should exist but got {resp.status_code}"
        assert resp.json() == data

    for record in non_exist_records:
        resp = client.get(f"{base_url}/{record}")
        assert (
            resp.status_code == 404
        ), f"Record {record} should not exist but got {resp.status_code}"



from __future__ import annotations

from dataclasses import dataclass
from typing import Annotated, Any, Callable

from graph.interface import BaseNode

from sera.libs.directed_computing_graph._fn_signature import FnSignature
from sera.libs.directed_computing_graph._type_conversion import UnitTypeConversion


class PartialFn:
    def __init__(self, fn: Callable, **kwargs):
        self.fn = fn
        self.default_args = kwargs
        self.signature = FnSignature.parse(fn)

        argnames = set(self.signature.argnames)
        for arg, val in self.default_args.items():
            if arg not in argnames:
                raise Exception(f"Argument {arg} is not in the function signature")
            self.signature.default_args[arg] = val

    def __call__(self, *args, **kwargs):
        return self.fn(*args, **kwargs)


ComputeFnId = Annotated[str, "ComputeFn Identifier"]
ComputeFn = PartialFn | Callable
NodeId = ComputeFnId


class DCGNode(BaseNode[NodeId]):
    id: NodeId
    func: ComputeFn

    def __init__(self, id: NodeId, func: ComputeFn):
        super().__init__(id)
        self.func = func
        self.signature = self.get_signature(self.func)
        self.type_conversions: list[UnitTypeConversion] = []
        self.required_args: list[str] = []
        self.required_context: list[str] = []
        self.required_context_default_args: dict[str, Any] = {}

    @staticmethod
    def get_signature(actor: ComputeFn) -> FnSignature:
        if isinstance(actor, PartialFn):
            return actor.signature
        else:
            return FnSignature.parse(actor)



from __future__ import annotations

import inspect
from dataclasses import dataclass
from typing import Any, Callable, get_args, get_origin, get_type_hints

from sera.misc import get_classpath


@dataclass
class FnSignature:
    return_type: type
    argnames: list[str]
    argtypes: list[type]
    default_args: dict[str, Any]  # Added this field to store default values
    is_async: bool = False

    @staticmethod
    def parse(func: Callable) -> FnSignature:
        sig = get_type_hints(func)
        argnames = list(sig.keys())[:-1]

        # Get the default values using inspect.signature
        inspect_sig = inspect.signature(func)
        defaults = {}
        for name, param in inspect_sig.parameters.items():
            if param.default is not inspect.Parameter.empty:
                defaults[name] = param.default

        try:
            return FnSignature(
                sig["return"],
                argnames,
                [sig[arg] for arg in argnames],
                defaults,  # Add the default values to the signature
                is_async=inspect.iscoroutinefunction(func),
            )
        except:
            print("Cannot figure out the signature of", func)
            print("The parsed signature is:", sig)
            raise


def type_to_string(_type: type) -> str:
    """Return a fully qualified type name"""
    origin = get_origin(_type)
    if origin is None:
        return get_classpath(_type)
    return (
        get_classpath(origin)
        + "["
        + ", ".join([get_classpath(arg) for arg in get_args(_type)])
        + "]"
    )



from __future__ import annotations

import os
from pathlib import Path
from pydoc import doc
from typing import Annotated

from codegen.models import DeferredVar, PredefinedFn, Program, expr, stmt
from loguru import logger
from sera.make.make_python_api import make_python_api
from sera.make.make_python_model import (
    make_python_data_model,
    make_python_enums,
    make_python_relational_model,
)
from sera.make.make_python_services import make_python_service_structure
from sera.make.make_typescript_model import (
    make_typescript_data_model,
    make_typescript_enum,
)
from sera.misc import Formatter
from sera.models import App, DataCollection, parse_schema
from sera.typing import Language


def make_config(app: App):
    """Make the configuration for the application."""
    if app.config.exists():
        logger.info("`{}` already exists. Skip generation.", app.config.path)
        return

    program = Program()
    program.import_("__future__.annotations", True)
    program.import_("os", False)
    program.import_("serde.yaml", False)
    program.import_("pathlib.Path", True)
    program.import_("sera.models.parse_schema", True)

    program.root(
        lambda ast: ast.if_(
            PredefinedFn.not_has_item(
                expr.ExprIdent("os.environ"), expr.ExprConstant("CFG_FILE")
            )
        )(
            lambda ast01: ast01.assign(
                DeferredVar.simple("CFG_FILE"),
                expr.ExprRawPython("Path(__file__).parent.parent / 'config.yml'"),
            ),
        ),
        lambda ast: ast.else_()(
            lambda ast01: ast01.assign(
                DeferredVar.simple("CFG_FILE"),
                expr.ExprRawPython('Path(os.environ["CFG_FILE"])'),
            ),
        ),
        lambda ast: ast.assign(
            DeferredVar.simple("cfg"),
            expr.ExprFuncCall(
                expr.ExprIdent("serde.yaml.deser"), [expr.ExprIdent("CFG_FILE")]
            ),
        ),
        stmt.LineBreak(),
        lambda ast: ast.assign(
            DeferredVar.simple("DB_CONNECTION"),
            expr.ExprIdent("cfg['db']['connection']"),
        ),
        lambda ast: ast.assign(
            DeferredVar.simple("DB_DEBUG"),
            expr.ExprIdent('os.environ.get("DB_DEBUG", "0") == "1"'),
        ),
        lambda ast: ast.assign(
            DeferredVar.simple("API_DEBUG"),
            expr.ExprIdent('os.environ.get("API_DEBUG", "0") == "1"'),
        ),
        stmt.LineBreak(),
        lambda ast: ast.assign(
            DeferredVar.simple("PKG_DIR"),
            expr.ExprRawPython("Path(__file__).parent"),
        ),
        lambda ast: ast.assign(
            DeferredVar.simple("schema"),
            expr.ExprFuncCall(
                expr.ExprIdent("parse_schema"),
                [
                    expr.ExprConstant(app.name),
                    PredefinedFn.list(
                        [
                            expr.ExprDivision(
                                expr.ExprIdent("PKG_DIR"),
                                expr.ExprConstant(
                                    os.path.relpath(path.absolute(), app.root.dir)
                                ),
                            )
                            for path in app.schema_files
                        ]
                    ),
                ],
            ),
        ),
    )

    app.config.write(program)


def make_app(
    app_dir: Annotated[
        Path,
        doc("Directory of the generated application"),
    ],
    schema_files: Annotated[
        list[Path],
        doc("YAML schema files. Multiple files are merged automatically"),
    ],
    api_collections: Annotated[
        list[str],
        doc("API collections to generate."),
    ],
    language: Annotated[
        Language,
        doc(
            "Language of the generated application. Currently only Python is supported."
        ),
    ] = Language.Python,
    referenced_schema: Annotated[
        list[str],
        doc(
            "Classes in the schema that are defined in different modules, used as references and thus should not be generated."
        ),
    ] = [],
):
    schema = parse_schema(app_dir.name, schema_files)

    app = App(app_dir.name, app_dir, schema_files, language)

    if language == Language.Python:
        # generate application configuration
        make_config(app)

        # generate models from schema
        # TODO: fix me, this is a hack to make the code work for referenced classes
        referenced_classes = {
            path.rsplit(".", 1)[1]: (parts := path.rsplit(".", 1))[0]
            + ".data."
            + parts[1]
            for path in referenced_schema
        }

        make_python_enums(schema, app.models.pkg("enums"), referenced_classes)
        make_python_data_model(schema, app.models.pkg("data"), referenced_classes)
        referenced_classes = {
            path.rsplit(".", 1)[1]: (parts := path.rsplit(".", 1))[0]
            + ".db."
            + parts[1]
            for path in referenced_schema
        }
        make_python_relational_model(
            schema,
            app.models.pkg("db"),
            app.models.pkg("data"),
            referenced_classes,
        )

        collections = [
            DataCollection(schema.classes[cname]) for cname in api_collections
        ]

        # generate API
        make_python_api(app, collections)

        # generate services
        make_python_service_structure(app, collections)
    elif language == Language.Typescript:
        make_typescript_enum(schema, app.models)
        make_typescript_data_model(schema, app.models)

    Formatter.get_instance().process()
    return app


if __name__ == "__main__":
    make_app(
        Path("/Volumes/research/workspace/libs/sera/tests/resources/myapp"),
        [
            Path(
                "/Volumes/research/workspace/libs/sera/tests/resources/schema/product.yml"
            )
        ],
        ["Product", "Category"],
    )



from sqlalchemy import (
    Column,
    ForeignKey,
    Integer,
    MetaData,
    String,
    Table,
    create_engine,
)
from sqlalchemy.schema import CreateTable

# Define your SQLAlchemy engine (dialect matters for SQL output)
# Using a specific dialect helps generate appropriate SQL
# engine = create_engine(
#     "postgresql+psycopg2://user:password@host:port/database", echo=False
# )
# Or for SQLite:
# engine = create_engine("sqlite:///:memory:")
# Or for MySQL:
# engine = create_engine("mysql+mysqlconnector://user:password@host:port/database")


metadata_obj = MetaData()

user_table = Table(
    "users",
    metadata_obj,
    Column("id", Integer, primary_key=True),
    Column("name", String(50)),
    Column("email", String(100), unique=True),
)

address_table = Table(
    "addresses",
    metadata_obj,
    Column("id", Integer, primary_key=True),
    Column("user_id", Integer, ForeignKey("users.id"), nullable=False),
    Column("street_name", String(100)),
    Column("city", String(50)),
)

# --- ORM Example ---
# from sqlalchemy.orm import declarative_base, Mapped, mapped_column
# Base = declarative_base()
# metadata_obj = Base.metadata
# class User(Base): # ... (define as above)
# class Address(Base): # ... (define as above)
# -------------------

print("--- Generating DDL for PostgreSQL ---")


def generate_schema_ddl(metadata, engine_dialect):
    for table in metadata.sorted_tables:
        # The CreateTable construct can be compiled to a string
        # specific to the dialect of the engine.
        create_table_ddl = CreateTable(table).compile(dialect=engine_dialect)
        print(str(create_table_ddl).strip() + ";\n")


from sqlalchemy.dialects import postgresql, sqlite

generate_schema_ddl(metadata_obj, postgresql.dialect())

# # Example with a different dialect (e.g., SQLite)
# # Note: You don't need a live connection for this, just the dialect.


# print("\n--- Generating DDL for SQLite ---")
# generate_schema_ddl(metadata_obj, sqlite.dialect())



from __future__ import annotations

import re
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Callable, Iterable, Optional, Sequence, Type, TypedDict, TypeVar

import serde.csv
import serde.json
from sqlalchemy import Engine, text
from sqlalchemy.orm import Session
from tqdm import tqdm

T = TypeVar("T")

TYPE_ALIASES = {"typing.List": "list", "typing.Dict": "dict", "typing.Set": "set"}

reserved_keywords = {
    "and",
    "or",
    "not",
    "is",
    "in",
    "if",
    "else",
    "elif",
    "for",
    "while",
    "def",
    "class",
    "return",
    "yield",
    "import",
    "from",
    "as",
    "with",
    "try",
    "except",
    "finally",
    "raise",
    "assert",
    "break",
    "continue",
    "pass",
    "del",
    "global",
    "nonlocal",
    "lambda",
    "async",
    "await",
    "True",
    "False",
    "None",
    "self",
}


def to_snake_case(camelcase: str) -> str:
    """Convert camelCase to snake_case."""
    snake = re.sub(r"([A-Z]+)([A-Z][a-z])", r"\1_\2", camelcase)
    snake = re.sub(r"([a-z0-9])([A-Z])", r"\1_\2", snake)
    return snake.lower()


def to_camel_case(snake: str) -> str:
    """Convert snake_case to camelCase."""
    components = snake.split("_")
    out = components[0] + "".join(x.title() for x in components[1:])
    # handle a corner case where the _ is added to the end of the string to avoid reserved keywords
    if snake.endswith("_") and snake[:-1] in reserved_keywords:
        out += "_"
    return out


def to_pascal_case(snake: str) -> str:
    """Convert snake_case to PascalCase."""
    components = snake.split("_")
    out = "".join(x.title() for x in components)
    # handle a corner case where the _ is added to the end of the string to avoid reserved keywords
    if snake.endswith("_") and snake[:-1] in reserved_keywords:
        out += "_"
    return out


def assert_isinstance(x: Any, cls: type[T]) -> T:
    if not isinstance(x, cls):
        raise Exception(f"{type(x)} doesn't match with {type(cls)}")
    return x


def assert_not_null(x: Optional[T]) -> T:
    assert x is not None
    return x


def filter_duplication(
    lst: Iterable[T], key_fn: Optional[Callable[[T], Any]] = None
) -> list[T]:
    keys = set()
    new_lst = []
    if key_fn is not None:
        for item in lst:
            k = key_fn(item)
            if k in keys:
                continue

            keys.add(k)
            new_lst.append(item)
    else:
        for k in lst:
            if k in keys:
                continue
            keys.add(k)
            new_lst.append(k)
    return new_lst


class LoadTableDataArgs(TypedDict, total=False):
    table: type
    tables: Sequence[type]
    file: Path
    files: Sequence[Path]
    file_deser: Callable[[Path], list[dict]]
    record_deser: Callable[[dict], Any | list[Any]]


def load_data(
    engine: Engine,
    create_db_and_tables: Callable[[], None],
    table_data: Sequence[LoadTableDataArgs],
    verbose: bool = False,
):
    """
    Load data into the database from specified CSV files.

    Args:
        engine: SQLAlchemy engine to connect to the database.
        create_db_and_tables: Function to create database and tables.
        table_files: List of tuples containing the class type and the corresponding CSV file path.
        table_desers: Dictionary mapping class types to their deserializer functions.
        verbose: If True, show progress bars during loading.
    """
    with Session(engine) as session:
        create_db_and_tables()

        for args in tqdm(table_data, disable=not verbose, desc="Loading data"):
            if "table" in args:
                tbls = [args["table"]]
            elif "tables" in args:
                tbls = args["tables"]
            else:
                raise ValueError("Either 'table' or 'tables' must be provided in args.")

            if "file" in args:
                assert isinstance(args["file"], Path), "File must be a Path object."
                files = [args["file"]]
            elif "files" in args:
                assert all(
                    isinstance(f, Path) for f in args["files"]
                ), "Files must be Path objects."
                files = args["files"]
            else:
                raise ValueError("Either 'file' or 'files' must be provided in args.")

            raw_records = []
            if "file_deser" not in args:
                for file in files:
                    if file.name.endswith(".csv"):
                        raw_records.extend(serde.csv.deser(file, deser_as_record=True))
                    elif file.name.endswith(".json"):
                        raw_records.extend(serde.json.deser(file))
                    else:
                        raise ValueError(f"Unsupported file format: {file.name}")
            else:
                for file in files:
                    raw_records.extend(args["file_deser"](file))

            deser = args["record_deser"]
            records = [deser(row) for row in raw_records]
            for r in tqdm(
                records,
                desc=f"load {', '.join(tbl.__name__ for tbl in tbls)}",
                disable=not verbose,
            ):
                if isinstance(r, Sequence):
                    for x in r:
                        session.merge(x)
                else:
                    session.merge(r)
            session.flush()

            # Reset the sequence for each table
            for tbl in tbls:
                # Check if the table has an auto-incrementing primary key
                if not hasattr(tbl, "__table__") or not tbl.__table__.primary_key:
                    continue

                pk_columns = tbl.__table__.primary_key.columns
                has_foreign_key = any(len(col.foreign_keys) > 0 for col in pk_columns)
                has_auto_increment = any(
                    col.autoincrement and col.type.python_type in (int,)
                    for col in pk_columns
                )
                if has_foreign_key or not has_auto_increment:
                    continue
                session.execute(
                    text(
                        f"SELECT setval('{tbl.__tablename__}_id_seq', (SELECT MAX(id) FROM \"{tbl.__tablename__}\"));"
                    )
                )
        session.commit()


def identity(x: T) -> T:
    """Identity function that returns the input unchanged."""
    return x


def get_classpath(type: Type | Callable) -> str:
    if type.__module__ == "builtins":
        return type.__qualname__

    if hasattr(type, "__qualname__"):
        return type.__module__ + "." + type.__qualname__

    # typically a class from the typing module
    if hasattr(type, "_name") and type._name is not None:
        path = type.__module__ + "." + type._name
        if path in TYPE_ALIASES:
            path = TYPE_ALIASES[path]
    elif hasattr(type, "__origin__") and hasattr(type.__origin__, "_name"):
        # found one case which is typing.Union
        path = type.__module__ + "." + type.__origin__._name
    else:
        raise NotImplementedError(type)

    return path



from __future__ import annotations

from typing import Sequence

from codegen.models import DeferredVar, Program, expr, stmt
from loguru import logger

from sera.misc import assert_not_null
from sera.models import App, DataCollection, Package


def make_python_service_structure(app: App, collections: Sequence[DataCollection]):
    """Make the basic structure for the service."""
    app.services.ensure_exists()

    for collection in collections:
        make_python_service(collection, app.services)


def make_python_service(collection: DataCollection, target_pkg: Package):
    app = target_pkg.app

    outmod = target_pkg.module(collection.get_pymodule_name())
    if outmod.exists():
        logger.info("`{}` already exists. Skip generation.", outmod.path)
        return

    # assuming the collection has only one class
    cls = collection.cls
    id_type = assert_not_null(cls.get_id_property()).datatype.get_python_type().type

    program = Program()
    program.import_("__future__.annotations", True)
    program.import_(
        app.models.db.path + f".{collection.get_pymodule_name()}.{collection.name}",
        True,
    )
    program.import_(app.config.path + f".schema", True)
    program.import_("sera.libs.base_service.BaseAsyncService", True)

    program.root(
        stmt.LineBreak(),
        lambda ast00: ast00.class_(
            collection.get_service_name(),
            [expr.ExprIdent(f"BaseAsyncService[{id_type}, {cls.name}]")],
        )(
            lambda ast01: ast01.func(
                "__init__",
                [
                    DeferredVar.simple("self"),
                ],
            )(
                lambda ast02: ast02.expr(
                    expr.ExprFuncCall(
                        expr.ExprIdent("super().__init__"),
                        [
                            expr.ExprRawPython(f"schema.classes['{cls.name}']"),
                            expr.ExprIdent(cls.name),
                        ],
                    )
                ),
            ),
        ),
    )
    outmod.write(program)



from __future__ import annotations

from typing import Sequence

from codegen.models import DeferredVar, ImportHelper, PredefinedFn, Program, expr, stmt
from loguru import logger

from sera.misc import assert_not_null, to_snake_case
from sera.models import App, DataCollection, Module, Package
from sera.typing import GLOBAL_IDENTS


def make_python_api(app: App, collections: Sequence[DataCollection]):
    """Make the basic structure for the API."""
    app.api.ensure_exists()
    app.api.pkg("routes").ensure_exists()

    # make routes
    routes: list[Module] = []
    for collection in collections:
        route = app.api.pkg("routes").pkg(collection.get_pymodule_name())

        controllers = []
        controllers.append(make_python_get_api(collection, route))
        controllers.append(make_python_get_by_id_api(collection, route))
        controllers.append(make_python_has_api(collection, route))
        controllers.append(make_python_create_api(collection, route))
        controllers.append(make_python_update_api(collection, route))

        routemod = route.module("route")
        if not routemod.exists():
            program = Program()
            program.import_("__future__.annotations", True)
            program.import_("litestar.Router", True)
            for get_route, get_route_fn in controllers:
                program.import_(get_route.path + "." + get_route_fn, True)

            program.root(
                stmt.LineBreak(),
                lambda ast: ast.assign(
                    DeferredVar.simple("router"),
                    expr.ExprFuncCall(
                        expr.ExprIdent("Router"),
                        [
                            PredefinedFn.keyword_assignment(
                                "path",
                                expr.ExprConstant(
                                    f"/api/{to_snake_case(collection.name).replace('_', '-')}"
                                ),
                            ),
                            PredefinedFn.keyword_assignment(
                                "route_handlers",
                                PredefinedFn.list(
                                    [
                                        expr.ExprIdent(get_route_fn)
                                        for get_route, get_route_fn in controllers
                                    ]
                                ),
                            ),
                        ],
                    ),
                ),
            )

            routemod.write(program)
        routes.append(routemod)

    # make the main entry point
    make_main(app.api, routes)


def make_main(target_pkg: Package, routes: Sequence[Module]):
    outmod = target_pkg.module("app")
    if outmod.exists():
        logger.info("`{}` already exists. Skip generation.", outmod.path)
        return

    program = Program()
    program.import_("__future__.annotations", True)
    program.import_("litestar.Litestar", True)
    for route in routes:
        program.import_(route.path, False)

    program.root(
        stmt.LineBreak(),
        lambda ast: ast.assign(
            DeferredVar.simple("app_routes"),
            PredefinedFn.list(
                [expr.ExprIdent(route.path + ".router") for route in routes]
            ),
        ),
        lambda ast: ast.assign(
            DeferredVar.simple("app"),
            expr.ExprFuncCall(
                expr.ExprIdent("Litestar"),
                [
                    PredefinedFn.keyword_assignment(
                        "route_handlers",
                        expr.ExprIdent("app_routes"),
                    )
                ],
            ),
        ),
    )

    outmod.write(program)


def make_python_get_api(
    collection: DataCollection, target_pkg: Package
) -> tuple[Module, str]:
    """Make an endpoint for querying resources"""
    app = target_pkg.app

    program = Program()
    import_helper = ImportHelper(program, GLOBAL_IDENTS)

    program.import_("__future__.annotations", True)
    program.import_("typing.Annotated", True)
    program.import_("litestar.get", True)
    program.import_("litestar.Request", True)
    program.import_("litestar.params.Parameter", True)
    program.import_(app.config.path + ".API_DEBUG", True)
    program.import_(
        app.services.path
        + f".{collection.get_pymodule_name()}.{collection.get_service_name()}",
        True,
    )
    program.import_(
        app.models.data.path + f".{collection.get_pymodule_name()}.{collection.name}",
        True,
    )
    program.import_("sera.libs.api_helper.parse_query", True)

    func_name = "get_"

    program.root(
        stmt.LineBreak(),
        lambda ast00: ast00.assign(
            DeferredVar.simple("QUERYABLE_FIELDS"),
            expr.ExprConstant(collection.get_queryable_fields()),
        ),
        stmt.PythonDecoratorStatement(
            expr.ExprFuncCall(
                expr.ExprIdent("get"),
                [
                    expr.ExprConstant("/"),
                ],
            )
        ),
        lambda ast10: ast10.func(
            func_name,
            [
                DeferredVar.simple(
                    "limit",
                    expr.ExprIdent(
                        'Annotated[int, Parameter(default=10, description="The maximum number of records to return")]'
                    ),
                ),
                DeferredVar.simple(
                    "offset",
                    type=expr.ExprIdent(
                        'Annotated[int, Parameter(default=0, description="The number of records to skip before returning results")]'
                    ),
                ),
                DeferredVar.simple(
                    "unique",
                    expr.ExprIdent(
                        'Annotated[bool, Parameter(default=False, description="Whether to return unique results only")]'
                    ),
                ),
                DeferredVar.simple(
                    "sorted_by",
                    expr.ExprIdent(
                        "Annotated[list[str], Parameter(default=tuple(), description=\"list of field names to sort by, prefix a field with '-' to sort that field in descending order\")]"
                    ),
                ),
                DeferredVar.simple(
                    "group_by",
                    expr.ExprIdent(
                        'Annotated[list[str], Parameter(default=tuple(), description="list of field names to group by")]'
                    ),
                ),
                DeferredVar.simple(
                    "fields",
                    expr.ExprIdent(
                        'Annotated[list[str], Parameter(default=tuple(), description="list of field names to include in the results")]'
                    ),
                ),
                DeferredVar.simple(
                    "request",
                    expr.ExprIdent("Request"),
                ),
                DeferredVar.simple(
                    "session",
                    import_helper.use("AsyncSession"),
                ),
            ],
            return_type=expr.ExprIdent(f"dict"),
            is_async=True,
        )(
            stmt.SingleExprStatement(
                expr.ExprConstant("Retrieving records matched a query")
            ),
            lambda ast100: ast100.assign(
                DeferredVar.simple("service"),
                expr.ExprFuncCall(
                    PredefinedFn.attr_getter(
                        expr.ExprIdent(collection.get_service_name()),
                        expr.ExprIdent("get_instance"),
                    ),
                    [],
                ),
            ),
            lambda ast101: ast101.assign(
                DeferredVar.simple("query", expr.ExprIdent("ServiceQuery")),
                expr.ExprFuncCall(
                    expr.ExprIdent("parse_query"),
                    [
                        expr.ExprIdent("request"),
                        expr.ExprIdent("QUERYABLE_FIELDS"),
                        PredefinedFn.keyword_assignment(
                            "debug",
                            expr.ExprIdent("API_DEBUG"),
                        ),
                    ],
                ),
            ),
            lambda ast102: ast102.assign(
                DeferredVar.simple("result"),
                expr.ExprAwait(
                    expr.ExprFuncCall(
                        PredefinedFn.attr_getter(
                            expr.ExprIdent("service"),
                            expr.ExprIdent("get"),
                        ),
                        [
                            expr.ExprIdent("query"),
                            PredefinedFn.keyword_assignment(
                                "limit", expr.ExprIdent("limit")
                            ),
                            PredefinedFn.keyword_assignment(
                                "offset", expr.ExprIdent("offset")
                            ),
                            PredefinedFn.keyword_assignment(
                                "unique", expr.ExprIdent("unique")
                            ),
                            PredefinedFn.keyword_assignment(
                                "sorted_by", expr.ExprIdent("sorted_by")
                            ),
                            PredefinedFn.keyword_assignment(
                                "group_by", expr.ExprIdent("group_by")
                            ),
                            PredefinedFn.keyword_assignment(
                                "fields", expr.ExprIdent("fields")
                            ),
                            PredefinedFn.keyword_assignment(
                                "session", expr.ExprIdent("session")
                            ),
                        ],
                    )
                ),
            ),
            lambda ast103: ast103.return_(
                PredefinedFn.dict(
                    [
                        (
                            PredefinedFn.attr_getter(
                                expr.ExprIdent(collection.name),
                                expr.ExprIdent("__name__"),
                            ),
                            PredefinedFn.map_list(
                                PredefinedFn.attr_getter(
                                    expr.ExprIdent("result"), expr.ExprIdent("records")
                                ),
                                lambda item: expr.ExprMethodCall(
                                    expr.ExprIdent(collection.name),
                                    "from_db",
                                    [item],
                                ),
                            ),
                        ),
                        (
                            expr.ExprConstant("total"),
                            PredefinedFn.attr_getter(
                                expr.ExprIdent("result"), expr.ExprIdent("total")
                            ),
                        ),
                    ]
                )
            ),
        ),
    )

    outmod = target_pkg.module("get")
    outmod.write(program)

    return outmod, func_name


def make_python_get_by_id_api(
    collection: DataCollection, target_pkg: Package
) -> tuple[Module, str]:
    """Make an endpoint for querying resource by id"""
    app = target_pkg.app

    program = Program()
    import_helper = ImportHelper(program, GLOBAL_IDENTS)

    program.import_("__future__.annotations", True)
    program.import_("litestar.get", True)
    program.import_("litestar.status_codes", True)
    program.import_("litestar.exceptions.HTTPException", True)
    program.import_(
        app.services.path
        + f".{collection.get_pymodule_name()}.{collection.get_service_name()}",
        True,
    )
    program.import_(
        app.models.data.path + f".{collection.get_pymodule_name()}.{collection.name}",
        True,
    )

    # assuming the collection has only one class
    cls = collection.cls
    id_type = assert_not_null(cls.get_id_property()).datatype.get_python_type().type

    func_name = "get_by_id"
    program.root(
        stmt.LineBreak(),
        stmt.PythonDecoratorStatement(
            expr.ExprFuncCall(
                expr.ExprIdent("get"),
                [
                    expr.ExprConstant("/{id:%s}" % id_type),
                ],
            )
        ),
        lambda ast10: ast10.func(
            func_name,
            [
                DeferredVar.simple(
                    "id",
                    expr.ExprIdent(id_type),
                ),
                DeferredVar.simple(
                    "session",
                    import_helper.use("AsyncSession"),
                ),
            ],
            return_type=expr.ExprIdent("dict"),
            is_async=True,
        )(
            stmt.SingleExprStatement(expr.ExprConstant("Retrieving record by id")),
            lambda ast100: ast100.assign(
                DeferredVar.simple("service"),
                expr.ExprFuncCall(
                    PredefinedFn.attr_getter(
                        expr.ExprIdent(collection.get_service_name()),
                        expr.ExprIdent("get_instance"),
                    ),
                    [],
                ),
            ),
            lambda ast11: ast11.assign(
                DeferredVar.simple("record"),
                expr.ExprAwait(
                    expr.ExprFuncCall(
                        expr.ExprIdent("service.get_by_id"),
                        [
                            expr.ExprIdent("id"),
                            expr.ExprIdent("session"),
                        ],
                    )
                ),
            ),
            lambda ast12: ast12.if_(PredefinedFn.is_null(expr.ExprIdent("record")))(
                lambda ast23: ast23.raise_exception(
                    expr.StandardExceptionExpr(
                        expr.ExprIdent("HTTPException"),
                        [
                            PredefinedFn.keyword_assignment(
                                "status_code",
                                expr.ExprIdent("status_codes.HTTP_404_NOT_FOUND"),
                            ),
                            PredefinedFn.keyword_assignment(
                                "detail",
                                expr.ExprIdent('f"Record with id {id} not found"'),
                            ),
                        ],
                    )
                )
            ),
            lambda ast13: ast13.return_(
                PredefinedFn.dict(
                    [
                        (
                            PredefinedFn.attr_getter(
                                expr.ExprIdent(cls.name), expr.ExprIdent("__name__")
                            ),
                            PredefinedFn.list(
                                [
                                    expr.ExprFuncCall(
                                        PredefinedFn.attr_getter(
                                            expr.ExprIdent(cls.name),
                                            expr.ExprIdent("from_db"),
                                        ),
                                        [expr.ExprIdent("record")],
                                    )
                                ]
                            ),
                        )
                    ]
                ),
            ),
        ),
    )

    outmod = target_pkg.module("get_by_id")
    outmod.write(program)

    return outmod, func_name


def make_python_has_api(
    collection: DataCollection, target_pkg: Package
) -> tuple[Module, str]:
    """Make an endpoint for querying resource by id"""
    app = target_pkg.app

    program = Program()
    import_helper = ImportHelper(program, GLOBAL_IDENTS)

    program.import_("__future__.annotations", True)
    program.import_("litestar.head", True)
    program.import_("litestar.status_codes", True)
    program.import_("litestar.exceptions.HTTPException", True)
    program.import_(
        app.services.path
        + f".{collection.get_pymodule_name()}.{collection.get_service_name()}",
        True,
    )

    # assuming the collection has only one class
    cls = collection.cls
    id_type = assert_not_null(cls.get_id_property()).datatype.get_python_type().type

    func_name = "has"
    program.root(
        stmt.LineBreak(),
        stmt.PythonDecoratorStatement(
            expr.ExprFuncCall(
                expr.ExprIdent("head"),
                [
                    expr.ExprConstant("/{id:%s}" % id_type),
                    PredefinedFn.keyword_assignment(
                        "status_code",
                        expr.ExprIdent("status_codes.HTTP_204_NO_CONTENT"),
                    ),
                ],
            )
        ),
        lambda ast10: ast10.func(
            func_name,
            [
                DeferredVar.simple(
                    "id",
                    expr.ExprIdent(id_type),
                ),
                DeferredVar.simple(
                    "session",
                    import_helper.use("AsyncSession"),
                ),
            ],
            return_type=expr.ExprConstant(None),
            is_async=True,
        )(
            stmt.SingleExprStatement(
                expr.ExprConstant("Checking if record exists by id")
            ),
            lambda ast100: ast100.assign(
                DeferredVar.simple("service"),
                expr.ExprFuncCall(
                    PredefinedFn.attr_getter(
                        expr.ExprIdent(collection.get_service_name()),
                        expr.ExprIdent("get_instance"),
                    ),
                    [],
                ),
            ),
            lambda ast11: ast11.assign(
                DeferredVar.simple("record_exist"),
                expr.ExprAwait(
                    expr.ExprFuncCall(
                        expr.ExprIdent("service.has_id"),
                        [
                            expr.ExprIdent("id"),
                            expr.ExprIdent("session"),
                        ],
                    )
                ),
            ),
            lambda ast12: ast12.if_(expr.ExprNegation(expr.ExprIdent("record_exist")))(
                lambda ast23: ast23.raise_exception(
                    expr.StandardExceptionExpr(
                        expr.ExprIdent("HTTPException"),
                        [
                            PredefinedFn.keyword_assignment(
                                "status_code",
                                expr.ExprIdent("status_codes.HTTP_404_NOT_FOUND"),
                            ),
                            PredefinedFn.keyword_assignment(
                                "detail",
                                expr.ExprIdent('f"Record with id {id} not found"'),
                            ),
                        ],
                    )
                )
            ),
            lambda ast13: ast13.return_(expr.ExprConstant(None)),
        ),
    )

    outmod = target_pkg.module("has")
    outmod.write(program)

    return outmod, func_name


def make_python_create_api(collection: DataCollection, target_pkg: Package):
    """Make an endpoint for creating a resource"""
    app = target_pkg.app

    program = Program()
    import_helper = ImportHelper(program, GLOBAL_IDENTS)

    program.import_("__future__.annotations", True)
    program.import_("litestar.post", True)
    program.import_(
        app.services.path
        + f".{collection.get_pymodule_name()}.{collection.get_service_name()}",
        True,
    )
    program.import_(
        app.models.data.path
        + f".{collection.get_pymodule_name()}.Create{collection.name}",
        True,
    )

    # assuming the collection has only one class
    cls = collection.cls
    is_on_create_update_props = any(
        prop.data.system_controlled is not None
        and prop.data.system_controlled.is_on_create_value_updated()
        for prop in cls.properties.values()
    )
    idprop = assert_not_null(cls.get_id_property())

    if is_on_create_update_props:
        program.import_("sera.libs.api_helper.SingleAutoUSCP", True)

    func_name = "create"

    program.root(
        stmt.LineBreak(),
        stmt.PythonDecoratorStatement(
            expr.ExprFuncCall(
                expr.ExprIdent("post"),
                [
                    expr.ExprConstant("/"),
                ]
                + (
                    [
                        PredefinedFn.keyword_assignment(
                            "dto",
                            PredefinedFn.item_getter(
                                expr.ExprIdent("SingleAutoUSCP"),
                                expr.ExprIdent(f"Create{cls.name}"),
                            ),
                        )
                    ]
                    if is_on_create_update_props
                    else []
                ),
            )
        ),
        lambda ast10: ast10.func(
            func_name,
            [
                DeferredVar.simple(
                    "data",
                    expr.ExprIdent(f"Create{cls.name}"),
                ),
                DeferredVar.simple(
                    "session",
                    import_helper.use("AsyncSession"),
                ),
            ],
            return_type=expr.ExprIdent(idprop.datatype.get_python_type().type),
            is_async=True,
        )(
            stmt.SingleExprStatement(expr.ExprConstant("Creating new record")),
            lambda ast100: ast100.assign(
                DeferredVar.simple("service"),
                expr.ExprFuncCall(
                    PredefinedFn.attr_getter(
                        expr.ExprIdent(collection.get_service_name()),
                        expr.ExprIdent("get_instance"),
                    ),
                    [],
                ),
            ),
            lambda ast13: ast13.return_(
                PredefinedFn.attr_getter(
                    expr.ExprAwait(
                        expr.ExprMethodCall(
                            expr.ExprIdent("service"),
                            "create",
                            [
                                expr.ExprMethodCall(
                                    expr.ExprIdent("data"), "to_db", []
                                ),
                                expr.ExprIdent("session"),
                            ],
                        )
                    ),
                    expr.ExprIdent(idprop.name),
                )
            ),
        ),
    )

    outmod = target_pkg.module("create")
    outmod.write(program)

    return outmod, func_name


def make_python_update_api(collection: DataCollection, target_pkg: Package):
    """Make an endpoint for updating resource"""
    app = target_pkg.app

    program = Program()
    import_helper = ImportHelper(program, GLOBAL_IDENTS)

    program.import_("__future__.annotations", True)
    program.import_("litestar.put", True)
    program.import_(
        app.services.path
        + f".{collection.get_pymodule_name()}.{collection.get_service_name()}",
        True,
    )
    program.import_(
        app.models.data.path
        + f".{collection.get_pymodule_name()}.Update{collection.name}",
        True,
    )

    # assuming the collection has only one class
    cls = collection.cls
    id_prop = assert_not_null(cls.get_id_property())
    id_type = id_prop.datatype.get_python_type().type

    is_on_update_update_props = any(
        prop.data.system_controlled is not None
        and prop.data.system_controlled.is_on_update_value_updated()
        for prop in cls.properties.values()
    )
    if is_on_update_update_props:
        program.import_("sera.libs.api_helper.SingleAutoUSCP", True)

    func_name = "update"

    program.root(
        stmt.LineBreak(),
        stmt.PythonDecoratorStatement(
            expr.ExprFuncCall(
                expr.ExprIdent("put"),
                [
                    expr.ExprConstant("/{id:%s}" % id_type),
                ]
                + (
                    [
                        PredefinedFn.keyword_assignment(
                            "dto",
                            PredefinedFn.item_getter(
                                expr.ExprIdent("SingleAutoUSCP"),
                                expr.ExprIdent(f"Update{cls.name}"),
                            ),
                        )
                    ]
                    if is_on_update_update_props
                    else []
                ),
            )
        ),
        lambda ast10: ast10.func(
            func_name,
            [
                DeferredVar.simple(
                    "id",
                    expr.ExprIdent(id_type),
                ),
                DeferredVar.simple(
                    "data",
                    expr.ExprIdent(f"Update{cls.name}"),
                ),
                DeferredVar.simple(
                    "session",
                    import_helper.use("AsyncSession"),
                ),
            ],
            return_type=expr.ExprIdent(id_prop.datatype.get_python_type().type),
            is_async=True,
        )(
            stmt.SingleExprStatement(expr.ExprConstant("Update an existing record")),
            lambda ast100: ast100.assign(
                DeferredVar.simple("service"),
                expr.ExprFuncCall(
                    PredefinedFn.attr_getter(
                        expr.ExprIdent(collection.get_service_name()),
                        expr.ExprIdent("get_instance"),
                    ),
                    [],
                ),
            ),
            stmt.SingleExprStatement(
                PredefinedFn.attr_setter(
                    expr.ExprIdent("data"),
                    expr.ExprIdent(id_prop.name),
                    expr.ExprIdent("id"),
                )
            ),
            lambda ast13: ast13.return_(
                PredefinedFn.attr_getter(
                    expr.ExprAwait(
                        expr.ExprMethodCall(
                            expr.ExprIdent("service"),
                            "update",
                            [
                                expr.ExprMethodCall(
                                    expr.ExprIdent("data"), "to_db", []
                                ),
                                expr.ExprIdent("session"),
                            ],
                        )
                    ),
                    expr.ExprIdent(id_prop.name),
                )
            ),
        ),
    )

    outmod = target_pkg.module("update")
    outmod.write(program)

    return outmod, func_name



from __future__ import annotations

from typing import Callable, Literal, Optional, Sequence

from codegen.models import (
    AST,
    DeferredVar,
    ImportHelper,
    PredefinedFn,
    Program,
    expr,
    stmt,
)

from sera.misc import (
    assert_isinstance,
    assert_not_null,
    filter_duplication,
    to_snake_case,
)
from sera.models import (
    Cardinality,
    Class,
    DataProperty,
    ObjectProperty,
    Package,
    PyTypeWithDep,
    Schema,
)
from sera.typing import GLOBAL_IDENTS, ObjectPath


def make_python_enums(
    schema: Schema,
    target_pkg: Package,
    reference_objects: dict[str, ObjectPath],
):
    """Make enums defined in the schema.

    Args:
        schema: The schema to generate the classes from.
        target_pkg: The package to write the enums to.
        reference_objects: A dictionary of objects to their references (e.g., the ones that are defined outside and used as referenced such as Tenant).
    """
    for enum in schema.enums.values():
        if enum.name in reference_objects:
            # skip enums that are defined in different apps
            continue

        program = Program()
        program.import_("__future__.annotations", True)
        program.import_("enum.Enum", True)

        enum_values = []
        for value in enum.values.values():
            enum_values.append(
                stmt.DefClassVarStatement(
                    name=value.name,
                    type=None,
                    value=expr.ExprConstant(value.value),
                )
            )

        program.root(
            stmt.LineBreak(),
            lambda ast: ast.class_(
                enum.name,
                (
                    [expr.ExprIdent("str")]
                    if enum.is_str_enum()
                    else [expr.ExprIdent("int")]
                )
                + [expr.ExprIdent("Enum")],
            )(*enum_values),
        )

        target_pkg.module(enum.get_pymodule_name()).write(program)


def make_python_data_model(
    schema: Schema, target_pkg: Package, reference_classes: dict[str, ObjectPath]
):
    """Generate public classes for the API from the schema.

    Args:
        schema: The schema to generate the classes from.
        target_pkg: The package to write the classes to.
        reference_classes: A dictionary of class names to their references (e.g., the ones that are defined outside and used as referenced such as Tenant).
    """
    app = target_pkg.app

    def from_db_type_conversion(
        record: expr.ExprIdent,
        prop: DataProperty | ObjectProperty,
        value_pass_as_args: Optional[expr.Expr] = None,
    ):
        """Convert the value from the database to the data model type.

        Args:
            record: The record to convert from.
            prop: The property to convert.
            value_pass_as_args: If provided, this value will be used instead of getting the value from the record. This is useful for functions that
                receive the column value as an argument, such as the `as_composite` function.
        """
        if value_pass_as_args is not None:
            value = value_pass_as_args
            assert record == expr.ExprIdent(
                ""
            ), "If value_pass_as_args is provided, record should not be used as a dummy value should be passed instead."
        else:
            value = PredefinedFn.attr_getter(record, expr.ExprIdent(prop.name))
        if isinstance(prop, ObjectProperty) and prop.target.db is not None:
            if prop.cardinality.is_star_to_many():
                value = PredefinedFn.map_list(
                    value,
                    lambda item: PredefinedFn.attr_getter(
                        item, expr.ExprIdent(prop.name + "_id")
                    ),
                )
            else:
                assert (
                    value_pass_as_args is None
                ), "Cannot use value_pass_as_args for a single object property."
                value = PredefinedFn.attr_getter(
                    record, expr.ExprIdent(prop.name + "_id")
                )

            target_idprop = assert_not_null(prop.target.get_id_property())
            conversion_fn = get_data_conversion(
                target_idprop.datatype.get_python_type().type,
                target_idprop.get_data_model_datatype().get_python_type().type,
            )
            value = conversion_fn(value)
        elif isinstance(prop, DataProperty) and prop.is_diff_data_model_datatype():
            value = get_data_conversion(
                prop.datatype.get_python_type().type,
                prop.get_data_model_datatype().get_python_type().type,
            )(value)

        return value

    def to_db_type_conversion(
        program: Program,
        slf: expr.ExprIdent,
        cls: Class,
        mode: Literal["create", "update"],
        prop: DataProperty | ObjectProperty,
    ):
        value = PredefinedFn.attr_getter(slf, expr.ExprIdent(prop.name))
        if isinstance(prop, ObjectProperty):
            if (
                prop.target.db is not None
                and prop.cardinality == Cardinality.MANY_TO_MANY
            ):
                # we have to use the associated object
                # if this isn't a many-to-many relationship, we only keep the id, so no need to convert to the type.
                AssociationTable = f"{cls.name}{prop.target.name}"
                program.import_(
                    app.models.db.path
                    + f".{to_snake_case(AssociationTable)}.{AssociationTable}",
                    True,
                )

                target_idprop = assert_not_null(prop.target.get_id_property())
                conversion_fn = get_data_conversion(
                    target_idprop.get_data_model_datatype().get_python_type().type,
                    target_idprop.datatype.get_python_type().type,
                )

                return PredefinedFn.map_list(
                    value,
                    lambda item: expr.ExprFuncCall(
                        expr.ExprIdent(AssociationTable),
                        [
                            PredefinedFn.keyword_assignment(
                                f"{prop.name}_id", conversion_fn(item)
                            )
                        ],
                    ),
                )
            elif prop.target.db is None:
                # if the target class is not in the database, we need to convert the value to the python type used in db.
                # if the cardinality is many-to-many, we need to convert each item in the list.
                if prop.cardinality.is_star_to_many():
                    value = PredefinedFn.map_list(
                        value, lambda item: expr.ExprMethodCall(item, "to_db", [])
                    )
                else:
                    value = expr.ExprMethodCall(value, "to_db", [])
        elif isinstance(prop, DataProperty) and prop.is_diff_data_model_datatype():
            # convert the value to the python type used in db
            converted_value = get_data_conversion(
                prop.get_data_model_datatype().get_python_type().type,
                prop.datatype.get_python_type().type,
            )(value)

            if mode == "update" and prop.data.is_private:
                # if the property is private and it's UNSET, we cannot transform it to the database type
                # and has to use the UNSET value (the update query will ignore this field)
                program.import_("sera.typing.UNSET", True)
                program.import_("sera.typing.is_set", True)
                value = expr.ExprTernary(
                    expr.ExprFuncCall(expr.ExprIdent("is_set"), [value]),
                    converted_value,
                    expr.ExprIdent("UNSET"),
                )
            else:
                value = converted_value
        return value

    def make_uscp_func(
        cls: Class,
        mode: Literal["create", "update"],
        ast: AST,
        ident_manager: ImportHelper,
    ):
        func = ast.func(
            "update_system_controlled_props",
            [
                DeferredVar.simple("self"),
                DeferredVar.simple(
                    "conn",
                    ident_manager.use("ASGIConnection"),
                ),
            ],
        )

        assign_user = False

        for prop in cls.properties.values():
            if prop.data.system_controlled is None:
                continue

            update_func = None
            if mode == "create":
                if prop.data.system_controlled.on_create_bypass is not None:
                    # by-pass the update function are handled later
                    continue

                if prop.data.system_controlled.is_on_create_value_updated():
                    update_func = (
                        prop.data.system_controlled.get_on_create_update_func()
                    )
            else:
                if prop.data.system_controlled.on_update_bypass is not None:
                    # by-pass the update function are handled later
                    continue
                if prop.data.system_controlled.is_on_update_value_updated():
                    update_func = (
                        prop.data.system_controlled.get_on_update_update_func()
                    )

            if update_func is None:
                continue

            if update_func.func == "getattr":
                if update_func.args[0] == "user":
                    if len(update_func.args) != 2:
                        raise NotImplementedError(
                            f"Unsupported update function: {update_func.func} with args {update_func.args}"
                        )
                    if not assign_user:
                        func(
                            stmt.AssignStatement(
                                expr.ExprIdent("user"),
                                PredefinedFn.item_getter(
                                    PredefinedFn.attr_getter(
                                        expr.ExprIdent("conn"),
                                        expr.ExprIdent("scope"),
                                    ),
                                    expr.ExprConstant("user"),
                                ),
                            )
                        )
                        assign_user = True
                    epr = PredefinedFn.attr_getter(
                        expr.ExprIdent("user"), expr.ExprIdent(update_func.args[1])
                    )
                elif update_func.args[0] == "self":
                    epr = PredefinedFn.attr_getter(
                        expr.ExprIdent("self"), expr.ExprIdent(update_func.args[1])
                    )
                else:
                    raise NotImplementedError(
                        f"Unsupported update function: {update_func.func} with args {update_func.args}"
                    )
            else:
                raise NotImplementedError(update_func.func)

            smt = stmt.AssignStatement(
                PredefinedFn.attr_getter(
                    expr.ExprIdent("self"), expr.ExprIdent(prop.name)
                ),
                epr,
            )
            func(smt)

        # handle the by-pass properties here
        for prop in cls.properties.values():
            if prop.data.system_controlled is None:
                continue

            update_func = None
            if mode == "create":
                if prop.data.system_controlled.on_create_bypass is None:
                    # non by-pass the update function are handled earlier
                    continue

                if prop.data.system_controlled.is_on_create_value_updated():
                    update_func = (
                        prop.data.system_controlled.get_on_create_update_func()
                    )
            else:
                if prop.data.system_controlled.on_update_bypass is None:
                    # non by-pass the update function are handled earlier
                    continue

                if prop.data.system_controlled.is_on_update_value_updated():
                    update_func = (
                        prop.data.system_controlled.get_on_update_update_func()
                    )

            if update_func is None:
                continue

            raise NotImplementedError("We haven't handled the by-pass properties yet.")

        func(
            stmt.AssignStatement(
                PredefinedFn.attr_getter(
                    expr.ExprIdent("self"), expr.ExprIdent("_is_scp_updated")
                ),
                expr.ExprConstant(True),
            )
        )

    def make_create(program: Program, cls: Class):
        program.import_("__future__.annotations", True)
        program.import_("msgspec", False)
        if cls.db is not None:
            # if the class is stored in the database, we need to import the database module
            program.import_(
                app.models.db.path + f".{cls.get_pymodule_name()}.{cls.name}",
                True,
                alias=f"{cls.name}DB",
            )

        ident_manager = ImportHelper(
            program,
            GLOBAL_IDENTS,
        )

        is_on_create_value_updated = any(
            prop.data.system_controlled is not None
            and prop.data.system_controlled.is_on_create_value_updated()
            for prop in cls.properties.values()
        )
        program.root.linebreak()
        cls_ast = program.root.class_(
            "Create" + cls.name,
            [expr.ExprIdent("msgspec.Struct"), expr.ExprIdent("kw_only=True")],
        )
        for prop in cls.properties.values():
            # Skip fields that are system-controlled (e.g., cached or derived fields)
            # and cannot be updated based on information parsed from the request.
            if (
                prop.data.system_controlled is not None
                and prop.data.system_controlled.is_on_create_ignored()
            ):
                continue

            if isinstance(prop, DataProperty):
                pytype = prop.get_data_model_datatype().get_python_type().clone()
                if len(prop.data.constraints) > 0:
                    # if the property has constraints, we need to figure out
                    program.import_("typing.Annotated", True)
                    if len(prop.data.constraints) == 1:
                        pytype.type = "Annotated[%s, %s]" % (
                            pytype.type,
                            prop.data.constraints[0].get_msgspec_constraint(),
                        )
                    else:
                        raise NotImplementedError(prop.data.constraints)

                if prop.is_optional:
                    pytype = pytype.as_optional_type()

                for dep in pytype.deps:
                    program.import_(dep, True)

                # private property are available for creating, but not for updating.
                # so we do not need to skip it.
                # if prop.data.is_private:
                #     program.import_("typing.Union", True)
                #     program.import_("sera.typing.UnsetType", True)
                #     program.import_("sera.typing.UNSET", True)
                #     pytype_type = f"Union[{pytype_type}, UnsetType]"

                prop_default_value = None
                # if prop.data.is_private:
                #     prop_default_value = expr.ExprIdent("UNSET")
                if prop.default_value is not None:
                    prop_default_value = expr.ExprConstant(prop.default_value)
                elif prop.default_factory is not None:
                    program.import_(prop.default_factory.pyfunc, True)
                    prop_default_value = expr.ExprFuncCall(
                        expr.ExprIdent("msgspec.field"),
                        [
                            PredefinedFn.keyword_assignment(
                                "default_factory",
                                expr.ExprIdent(prop.default_factory.pyfunc),
                            )
                        ],
                    )

                cls_ast(
                    stmt.DefClassVarStatement(
                        prop.name, pytype.type, prop_default_value
                    )
                )
            elif isinstance(prop, ObjectProperty):
                if prop.target.db is not None:
                    # if the target class is in the database, we expect the user to pass the foreign key for it.
                    pytype = (
                        assert_not_null(prop.target.get_id_property())
                        .get_data_model_datatype()
                        .get_python_type()
                    )
                else:
                    pytype = PyTypeWithDep(
                        f"Create{prop.target.name}",
                        [
                            f"{target_pkg.module(prop.target.get_pymodule_name()).path}.Create{prop.target.name}"
                        ],
                    )

                if prop.cardinality.is_star_to_many():
                    pytype = pytype.as_list_type()
                elif prop.is_optional:
                    pytype = pytype.as_optional_type()

                for dep in pytype.deps:
                    program.import_(dep, True)

                cls_ast(stmt.DefClassVarStatement(prop.name, pytype.type))

        if is_on_create_value_updated:
            program.import_("typing.Optional", True)
            program.import_("sera.typing.is_set", True)
            cls_ast(
                stmt.Comment(
                    "A marker to indicate that the system-controlled properties are updated"
                ),
                stmt.DefClassVarStatement(
                    "_is_scp_updated", "bool", expr.ExprConstant(False)
                ),
                stmt.LineBreak(),
                lambda ast: ast.func(
                    "__post_init__",
                    [
                        DeferredVar.simple("self"),
                    ],
                )(
                    stmt.AssignStatement(
                        PredefinedFn.attr_getter(
                            expr.ExprIdent("self"), expr.ExprIdent("_is_scp_updated")
                        ),
                        expr.ExprConstant(False),
                    ),
                ),
                stmt.LineBreak(),
                lambda ast: make_uscp_func(cls, "create", ast, ident_manager),
            )

        cls_ast(
            stmt.LineBreak(),
            lambda ast00: ast00.func(
                "to_db",
                [
                    DeferredVar.simple("self"),
                ],
                return_type=expr.ExprIdent(
                    f"{cls.name}DB" if cls.db is not None else cls.name
                ),
            )(
                (
                    stmt.AssertionStatement(
                        PredefinedFn.attr_getter(
                            expr.ExprIdent("self"),
                            expr.ExprIdent("_is_scp_updated"),
                        ),
                        expr.ExprConstant(
                            "The model data must be verified before converting to db model"
                        ),
                    )
                    if is_on_create_value_updated
                    else None
                ),
                lambda ast10: ast10.return_(
                    expr.ExprFuncCall(
                        expr.ExprIdent(
                            f"{cls.name}DB" if cls.db is not None else cls.name
                        ),
                        [
                            (
                                ident_manager.use("UNSET")
                                if prop.data.system_controlled is not None
                                and prop.data.system_controlled.is_on_create_ignored()
                                else to_db_type_conversion(
                                    program, expr.ExprIdent("self"), cls, "create", prop
                                )
                            )
                            for prop in cls.properties.values()
                        ],
                    )
                ),
            ),
        )

    def make_update(program: Program, cls: Class):
        program.import_("__future__.annotations", True)
        program.import_("msgspec", False)
        if cls.db is not None:
            # if the class is stored in the database, we need to import the database module
            program.import_(
                app.models.db.path + f".{cls.get_pymodule_name()}.{cls.name}",
                True,
                alias=f"{cls.name}DB",
            )

        ident_manager = ImportHelper(
            program,
            GLOBAL_IDENTS,
        )

        # property that normal users cannot set, but super users can
        is_on_update_value_updated = any(
            prop.data.system_controlled is not None
            and prop.data.system_controlled.is_on_update_value_updated()
            for prop in cls.properties.values()
        )

        program.root.linebreak()
        cls_ast = program.root.class_(
            "Update" + cls.name,
            [expr.ExprIdent("msgspec.Struct"), expr.ExprIdent("kw_only=True")],
        )
        for prop in cls.properties.values():
            # Skip fields that are system-controlled (e.g., cached or derived fields)
            # and cannot be updated based on information parsed from the request.
            if (
                prop.data.system_controlled is not None
                and prop.data.system_controlled.is_on_update_ignored()
            ):
                continue

            if isinstance(prop, DataProperty):
                pytype = prop.get_data_model_datatype().get_python_type().clone()

                if len(prop.data.constraints) > 0:
                    # if the property has constraints, we need to figure out
                    program.import_("typing.Annotated", True)
                    if len(prop.data.constraints) == 1:
                        pytype.type = "Annotated[%s, %s]" % (
                            pytype.type,
                            prop.data.constraints[0].get_msgspec_constraint(),
                        )
                    else:
                        raise NotImplementedError(prop.data.constraints)

                if prop.is_optional:
                    pytype = pytype.as_optional_type()

                for dep in pytype.deps:
                    program.import_(dep, True)

                if prop.data.is_private:
                    program.import_("typing.Union", True)
                    program.import_("sera.typing.UnsetType", True)
                    program.import_("sera.typing.UNSET", True)
                    pytype.type = f"Union[{pytype.type}, UnsetType]"

                prop_default_value = None
                if prop.data.is_private:
                    prop_default_value = expr.ExprIdent("UNSET")
                elif prop.default_value is not None:
                    prop_default_value = expr.ExprConstant(prop.default_value)
                elif prop.default_factory is not None:
                    program.import_(prop.default_factory.pyfunc, True)
                    prop_default_value = expr.ExprFuncCall(
                        expr.ExprIdent("msgspec.field"),
                        [
                            PredefinedFn.keyword_assignment(
                                "default_factory",
                                expr.ExprIdent(prop.default_factory.pyfunc),
                            )
                        ],
                    )

                cls_ast(
                    stmt.DefClassVarStatement(
                        prop.name, pytype.type, prop_default_value
                    )
                )
            elif isinstance(prop, ObjectProperty):
                if prop.target.db is not None:
                    # if the target class is in the database, we expect the user to pass the foreign key for it.
                    pytype = (
                        assert_not_null(prop.target.get_id_property())
                        .get_data_model_datatype()
                        .get_python_type()
                    )
                else:
                    pytype = PyTypeWithDep(
                        f"Update{prop.target.name}",
                        [
                            f"{target_pkg.module(prop.target.get_pymodule_name()).path}.Update{prop.target.name}"
                        ],
                    )

                if prop.cardinality.is_star_to_many():
                    pytype = pytype.as_list_type()
                elif prop.is_optional:
                    pytype = pytype.as_optional_type()

                for dep in pytype.deps:
                    program.import_(dep, True)

                cls_ast(stmt.DefClassVarStatement(prop.name, pytype.type))

        if is_on_update_value_updated:
            program.import_("typing.Optional", True)
            program.import_("sera.typing.is_set", True)
            cls_ast(
                stmt.Comment(
                    "A marker to indicate that the system-controlled properties are updated"
                ),
                stmt.DefClassVarStatement(
                    "_is_scp_updated", "bool", expr.ExprConstant(False)
                ),
                stmt.LineBreak(),
                lambda ast: ast.func(
                    "__post_init__",
                    [
                        DeferredVar.simple("self"),
                    ],
                )(
                    stmt.AssignStatement(
                        PredefinedFn.attr_getter(
                            expr.ExprIdent("self"), expr.ExprIdent("_is_scp_updated")
                        ),
                        expr.ExprConstant(False),
                    ),
                ),
                stmt.LineBreak(),
                lambda ast: make_uscp_func(cls, "update", ast, ident_manager),
            )

        cls_ast(
            stmt.LineBreak(),
            lambda ast00: ast00.func(
                "to_db",
                [
                    DeferredVar.simple("self"),
                ],
                return_type=expr.ExprIdent(
                    f"{cls.name}DB" if cls.db is not None else cls.name
                ),
            )(
                (
                    stmt.AssertionStatement(
                        expr.ExprLogicalAnd(
                            [
                                expr.ExprFuncCall(
                                    expr.ExprIdent("is_set"),
                                    [
                                        PredefinedFn.attr_getter(
                                            expr.ExprIdent("self"),
                                            expr.ExprIdent(prop.name),
                                        )
                                    ],
                                )
                                for prop in cls.properties.values()
                                if prop.data.system_controlled is not None
                                and prop.data.system_controlled.is_on_update_value_updated()
                            ]
                        ),
                        expr.ExprConstant(
                            "The model data must be verified before converting to db model"
                        ),
                    )
                    if is_on_update_value_updated
                    else None
                ),
                lambda ast10: ast10.return_(
                    expr.ExprFuncCall(
                        expr.ExprIdent(
                            f"{cls.name}DB" if cls.db is not None else cls.name
                        ),
                        [
                            (
                                ident_manager.use("UNSET")
                                if prop.data.system_controlled is not None
                                and prop.data.system_controlled.is_on_update_ignored()
                                else to_db_type_conversion(
                                    program, expr.ExprIdent("self"), cls, "update", prop
                                )
                            )
                            for prop in cls.properties.values()
                        ],
                    )
                ),
            ),
        )

    def make_normal(program: Program, cls: Class):
        if not cls.is_public:
            # skip classes that are not public
            return

        program.import_("__future__.annotations", True)
        program.import_("msgspec", False)

        ident_manager = ImportHelper(
            program,
            GLOBAL_IDENTS,
        )

        if cls.db is not None:
            # if the class is stored in the database, we need to import the database module
            program.import_(
                app.models.db.path + f".{cls.get_pymodule_name()}.{cls.name}",
                True,
                alias=f"{cls.name}DB",
            )

        cls_ast = program.root.class_(cls.name, [expr.ExprIdent("msgspec.Struct")])
        for prop in cls.properties.values():
            if prop.data.is_private:
                # skip private fields as this is for APIs exchange
                continue

            if isinstance(prop, DataProperty):
                pytype = prop.get_data_model_datatype().get_python_type()
                if prop.is_optional:
                    pytype = pytype.as_optional_type()

                for dep in pytype.deps:
                    program.import_(dep, True)

                cls_ast(stmt.DefClassVarStatement(prop.name, pytype.type))
            elif isinstance(prop, ObjectProperty):
                if prop.target.db is not None:
                    pytype = (
                        assert_not_null(prop.target.get_id_property())
                        .get_data_model_datatype()
                        .get_python_type()
                    )
                else:
                    pytype = PyTypeWithDep(
                        prop.target.name,
                        [
                            f"{target_pkg.module(prop.target.get_pymodule_name()).path}.{prop.target.name}"
                        ],
                    )

                if prop.cardinality.is_star_to_many():
                    pytype = pytype.as_list_type()
                elif prop.is_optional:
                    pytype = pytype.as_optional_type()

                for dep in pytype.deps:
                    program.import_(dep, True)

                cls_ast(stmt.DefClassVarStatement(prop.name, pytype.type))

        cls_ast(
            stmt.LineBreak(),
            (
                stmt.PythonDecoratorStatement(
                    expr.ExprFuncCall(expr.ExprIdent("classmethod"), [])
                )
                if cls.db is not None
                else None
            ),
            lambda ast00: (
                ast00.func(
                    "from_db",
                    [
                        DeferredVar.simple("cls"),
                        DeferredVar.simple("record", expr.ExprIdent(f"{cls.name}DB")),
                    ],
                )(
                    lambda ast10: ast10.return_(
                        expr.ExprFuncCall(
                            expr.ExprIdent("cls"),
                            [
                                from_db_type_conversion(expr.ExprIdent("record"), prop)
                                for prop in cls.properties.values()
                                if not prop.data.is_private
                            ],
                        )
                    )
                )
                if cls.db is not None
                else None
            ),
        )

        if cls.db is None:
            as_composite_args = [DeferredVar.simple("cls")]
            as_composite_null_condition = []
            for prop in cls.properties.values():
                assert (
                    not prop.data.is_private
                ), f"Embedded classes should not have private properties: {cls.name}.{prop.name}"
                as_composite_args.append(DeferredVar.simple(prop.name))
                as_composite_null_condition.append(
                    expr.ExprIs(expr.ExprIdent(prop.name), expr.ExprConstant(None))
                )

            # For simplicity, we assume that this embedded class can be used in a nullable field (check if all properties are None and return None).
            # However, we could be more efficient by checking if there are any other classes that use this class as a composite and are non-optional,
            # and eliminate the None check because we know that the class will always be re-created.
            cls_ast(
                stmt.LineBreak(),
                stmt.PythonDecoratorStatement(
                    expr.ExprFuncCall(expr.ExprIdent("classmethod"), [])
                ),
                lambda ast: ast.func(
                    "as_composite",
                    vars=as_composite_args,
                    return_type=PredefinedFn.item_getter(
                        ident_manager.use("Optional"), expr.ExprIdent(cls.name)
                    ),
                    comment="Create an embedded instance from the embedded columns in the database table. If all properties of this embedded class are None (indicating that the parent field is None), then this function will return None.",
                )(
                    lambda ast_l1: ast_l1.if_(
                        expr.ExprLogicalAnd(as_composite_null_condition)
                    )(lambda ast_l2: ast_l2.return_(expr.ExprConstant(None))),
                    lambda ast_l1: ast_l1.return_(
                        expr.ExprFuncCall(
                            expr.ExprIdent("cls"),
                            [
                                from_db_type_conversion(
                                    expr.ExprIdent(""), prop, expr.ExprIdent(prop.name)
                                )
                                for prop in cls.properties.values()
                            ],
                        )
                    ),
                ),
                stmt.LineBreak(),
                lambda ast: ast.func(
                    "__composite_values__",
                    [
                        DeferredVar.simple("self"),
                    ],
                    return_type=expr.ExprIdent("tuple"),
                    comment="Return the values of the properties of this embedded class as a tuple. This is used to create a composite object in SQLAlchemy.",
                )(
                    lambda ast_l1: ast_l1.return_(
                        PredefinedFn.tuple(
                            [
                                to_db_type_conversion(
                                    program, expr.ExprIdent("self"), cls, "create", prop
                                )
                                for prop in cls.properties.values()
                            ]
                        )
                    )
                ),
            )

    for cls in schema.topological_sort():
        if cls.name in reference_classes:
            continue

        program = Program()
        make_create(program, cls)
        program.root.linebreak()
        make_update(program, cls)
        program.root.linebreak()
        make_normal(program, cls)
        target_pkg.module(cls.get_pymodule_name()).write(program)


def make_python_relational_model(
    schema: Schema,
    target_pkg: Package,
    target_data_pkg: Package,
    reference_classes: dict[str, ObjectPath],
):
    """Make python classes for relational database using SQLAlchemy.

    The new classes is going be compatible with SQLAlchemy 2.

    Args:
        schema: The schema to generate the classes from.
        target_pkg: The package to write the classes to.
        target_data_pkg: The package to write the data classes to.
        reference_classes: A dictionary of class names to their references (e.g., the ones that are defined outside and used as referenced such as Tenant).
    """
    app = target_pkg.app

    def get_property_name(prop: DataProperty | ObjectProperty):
        if isinstance(prop, ObjectProperty):
            if prop.target.db is not None:
                return f"{prop.name}_id"
        return prop.name

    def make_base(custom_types: Sequence[ObjectProperty]):
        """Make a base class for our database."""
        program = Program()
        program.import_("__future__.annotations", True)
        program.import_("sera.libs.base_orm.BaseORM", True)
        program.import_("sera.libs.base_orm.create_engine", True)
        program.import_("sera.libs.base_orm.create_async_engine", True)
        program.import_("sqlalchemy.orm.DeclarativeBase", True)
        program.import_("sqlalchemy.orm.Session", True)
        program.import_("sqlalchemy.ext.asyncio.AsyncSession", True)
        program.import_("sqlalchemy.text", True)

        # assume configuration for the app at the top level
        program.import_(f"{app.config.path}.DB_CONNECTION", True)
        program.import_(f"{app.config.path}.DB_DEBUG", True)
        program.import_(f"contextlib.contextmanager", True)

        program.root.linebreak()

        type_map = []
        for custom_type in custom_types:
            program.import_(
                f"{target_data_pkg.module(custom_type.target.get_pymodule_name()).path}.{custom_type.target.name}",
                is_import_attr=True,
            )

            if custom_type.cardinality.is_star_to_many():
                if custom_type.is_map:
                    program.import_("sera.libs.base_orm.DictDataclassType", True)
                    type = f"dict[str, {custom_type.target.name}]"
                    maptype = f"DictDataclassType({custom_type.target.name})"
                else:
                    program.import_("sera.libs.base_orm.ListDataclassType", True)
                    type = f"list[{custom_type.target.name}]"
                    maptype = f"ListDataclassType({custom_type.target.name})"
            else:
                program.import_("sera.libs.base_orm.DataclassType", True)
                type = custom_type.target.name
                maptype = f"DataclassType({custom_type.target.name})"

            if custom_type.is_optional:
                program.import_("typing.Optional", True)
                type = f"Optional[{type}]"

            type_map.append((expr.ExprIdent(type), expr.ExprIdent(maptype)))

        program.root.class_(
            "Base", [expr.ExprIdent("DeclarativeBase"), expr.ExprIdent("BaseORM")]
        )(
            stmt.DefClassVarStatement(
                "type_annotation_map", "dict", PredefinedFn.dict(type_map)
            ),
            return_self=True,
        )

        program.root.linebreak()
        program.root.assign(
            DeferredVar.simple("engine"),
            expr.ExprFuncCall(
                expr.ExprIdent("create_engine"),
                [
                    expr.ExprIdent("DB_CONNECTION"),
                    PredefinedFn.keyword_assignment("echo", expr.ExprIdent("DB_DEBUG")),
                ],
            ),
        )
        program.root.assign(
            DeferredVar.simple("async_engine"),
            expr.ExprFuncCall(
                expr.ExprIdent("create_async_engine"),
                [
                    expr.ExprIdent("DB_CONNECTION"),
                    PredefinedFn.keyword_assignment("echo", expr.ExprIdent("DB_DEBUG")),
                ],
            ),
        )

        program.root.linebreak()
        program.root.func("create_db_and_tables", [])(
            stmt.PythonStatement("Base.metadata.create_all(engine)"),
        )

        program.root.linebreak()
        program.root.func("get_async_session", [], is_async=True)(
            lambda ast: ast.python_stmt(
                "async with AsyncSession(async_engine, expire_on_commit=False) as session:"
            )(
                lambda ast_l1: ast_l1.try_()(stmt.PythonStatement("yield session")),
                lambda ast_l1: ast_l1.catch()(
                    stmt.SingleExprStatement(
                        expr.ExprAwait(
                            expr.ExprFuncCall(
                                PredefinedFn.attr_getter(
                                    expr.ExprIdent("session"),
                                    expr.ExprIdent("rollback"),
                                ),
                                [],
                            )
                        )
                    ),
                    stmt.PythonStatement("raise"),
                ),
                lambda ast_l1: ast_l1.else_()(
                    stmt.SingleExprStatement(
                        expr.ExprAwait(
                            expr.ExprFuncCall(
                                PredefinedFn.attr_getter(
                                    expr.ExprIdent("session"), expr.ExprIdent("execute")
                                ),
                                [
                                    expr.ExprFuncCall(
                                        expr.ExprIdent("text"),
                                        [expr.ExprConstant("RESET ROLE;")],
                                    )
                                ],
                            )
                        )
                    ),
                    stmt.SingleExprStatement(
                        expr.ExprAwait(
                            expr.ExprFuncCall(
                                PredefinedFn.attr_getter(
                                    expr.ExprIdent("session"), expr.ExprIdent("commit")
                                ),
                                [],
                            )
                        )
                    ),
                ),
            )
        )

        program.root.linebreak()
        program.root.python_stmt("@contextmanager")
        program.root.func("get_session", [])(
            lambda ast: ast.python_stmt(
                "with Session(engine, expire_on_commit=False) as session:"
            )(
                lambda ast_l1: ast_l1.try_()(stmt.PythonStatement("yield session")),
                lambda ast_l1: ast_l1.catch()(
                    stmt.SingleExprStatement(
                        expr.ExprFuncCall(
                            PredefinedFn.attr_getter(
                                expr.ExprIdent("session"), expr.ExprIdent("rollback")
                            ),
                            [],
                        )
                    ),
                    stmt.PythonStatement("raise"),
                ),
                lambda ast_l1: ast_l1.else_()(
                    stmt.SingleExprStatement(
                        expr.ExprFuncCall(
                            PredefinedFn.attr_getter(
                                expr.ExprIdent("session"), expr.ExprIdent("execute")
                            ),
                            [
                                expr.ExprFuncCall(
                                    expr.ExprIdent("text"),
                                    [expr.ExprConstant("RESET ROLE;")],
                                )
                            ],
                        )
                    ),
                    stmt.SingleExprStatement(
                        expr.ExprFuncCall(
                            PredefinedFn.attr_getter(
                                expr.ExprIdent("session"), expr.ExprIdent("commit")
                            ),
                            [],
                        )
                    ),
                ),
            )
        )

        target_pkg.module("base").write(program)

    def make_orm(cls: Class):
        if cls.db is None or cls.name in reference_classes:
            # skip classes that are not stored in the database
            return

        program = Program()
        program.import_("__future__.annotations", True)
        program.import_("sqlalchemy.orm.MappedAsDataclass", True)
        program.import_("sqlalchemy.orm.mapped_column", True)
        program.import_("sqlalchemy.orm.Mapped", True)
        program.import_(f"{target_pkg.path}.base.Base", True)

        ident_manager = ImportHelper(
            program,
            GLOBAL_IDENTS,
        )

        index_stmts = []
        if len(cls.db.indices) > 0:
            program.import_("sqlalchemy.Index", True)
            index_stmts.append(
                stmt.DefClassVarStatement(
                    "_table_args__",
                    None,
                    PredefinedFn.tuple(
                        [
                            expr.ExprFuncCall(
                                expr.ExprIdent("Index"),
                                [expr.ExprConstant(index.name)]
                                + [
                                    expr.ExprConstant(
                                        get_property_name(cls.properties[prop])
                                    )
                                    for prop in index.columns
                                ]
                                + (
                                    [
                                        PredefinedFn.keyword_assignment(
                                            "unique", expr.ExprConstant(index.unique)
                                        )
                                    ]
                                    if index.unique
                                    else []
                                ),
                            )
                            for index in cls.db.indices
                        ]
                    ),
                )
            )

        cls_ast = program.root.class_(
            cls.name, [expr.ExprIdent("MappedAsDataclass"), expr.ExprIdent("Base")]
        )
        cls_ast(
            stmt.DefClassVarStatement(
                "__tablename__",
                type=None,
                value=expr.ExprConstant(cls.db.table_name),
            ),
            *index_stmts,
            stmt.LineBreak(),
        )

        for prop in cls.properties.values():
            if prop.db is None:
                # skip properties that are not stored in the database
                continue

            if isinstance(prop, DataProperty):
                sqltype = prop.datatype.get_sqlalchemy_type()
                for dep in sqltype.deps:
                    program.import_(dep, True)

                propname = prop.name

                if prop.is_optional:
                    program.import_("typing.Optional", True)
                    proptype = f"Mapped[Optional[{sqltype.mapped_pytype}]]"
                else:
                    proptype = f"Mapped[{sqltype.mapped_pytype}]"

                propvalargs: list[expr.Expr] = [expr.ExprIdent(sqltype.type)]
                if prop.db.foreign_key is not None:
                    assert (
                        prop.db.foreign_key.db is not None
                    ), f"Foreign key {prop.db.foreign_key.name} must have a database mapping"
                    foreign_key_idprop = prop.db.foreign_key.get_id_property()
                    assert (
                        foreign_key_idprop is not None
                    ), f"Foreign key {prop.db.foreign_key.name} must have an id property"
                    propvalargs.append(
                        expr.ExprFuncCall(
                            ident_manager.use("ForeignKey"),
                            [
                                expr.ExprConstant(
                                    f"{prop.db.foreign_key.db.table_name}.{foreign_key_idprop.name}"
                                ),
                                PredefinedFn.keyword_assignment(
                                    "ondelete",
                                    expr.ExprConstant("CASCADE"),
                                ),
                                PredefinedFn.keyword_assignment(
                                    "onupdate",
                                    expr.ExprConstant("CASCADE"),
                                ),
                            ],
                        )
                    )
                if prop.db.is_primary_key:
                    propvalargs.append(
                        PredefinedFn.keyword_assignment(
                            "primary_key", expr.ExprConstant(True)
                        )
                    )
                    if prop.db.is_auto_increment:
                        propvalargs.append(
                            PredefinedFn.keyword_assignment(
                                "autoincrement", expr.ExprConstant("auto")
                            )
                        )
                else:
                    if prop.db.is_unique:
                        propvalargs.append(
                            PredefinedFn.keyword_assignment(
                                "unique", expr.ExprConstant(True)
                            )
                        )
                    elif prop.db.is_indexed:
                        propvalargs.append(
                            PredefinedFn.keyword_assignment(
                                "index", expr.ExprConstant(True)
                            )
                        )
                if prop.is_optional:
                    propvalargs.append(
                        PredefinedFn.keyword_assignment(
                            "nullable", expr.ExprConstant(True)
                        )
                    )
                propval = expr.ExprFuncCall(
                    expr.ExprIdent("mapped_column"), propvalargs
                )
                cls_ast(stmt.DefClassVarStatement(propname, proptype, propval))
            else:
                assert isinstance(prop, ObjectProperty)
                make_python_relational_object_property(
                    program=program,
                    target_pkg=target_pkg,
                    target_data_pkg=target_data_pkg,
                    cls_ast=cls_ast,
                    cls=cls,
                    prop=prop,
                    custom_types=custom_types,
                )

        target_pkg.module(cls.get_pymodule_name()).write(program)

    custom_types: list[ObjectProperty] = []

    for cls in schema.topological_sort():
        make_orm(cls)

    # make a base class that implements the mapping for custom types
    custom_types = filter_duplication(
        custom_types, lambda p: (p.target.name, p.cardinality, p.is_optional, p.is_map)
    )
    make_base(custom_types)


def make_python_relational_object_property(
    program: Program,
    target_pkg: Package,
    target_data_pkg: Package,
    cls_ast: AST,
    cls: Class,
    prop: ObjectProperty,
    custom_types: list[ObjectProperty],
):
    assert prop.db is not None
    if prop.target.db is not None:
        # if the target class is in the database, we generate a foreign key for it.
        program.import_("sqlalchemy.ForeignKey", True)

        if prop.cardinality == Cardinality.MANY_TO_MANY:
            make_python_relational_object_property_many_to_many(
                program, cls_ast, target_pkg, cls, prop
            )
            return

        if prop.cardinality.is_star_to_many():
            raise NotImplementedError((cls.name, prop.name))

        # we store this class in the database
        propname = f"{prop.name}_id"
        idprop = prop.target.get_id_property()
        assert idprop is not None
        idprop_pytype = idprop.datatype.get_sqlalchemy_type()
        for dep in idprop_pytype.deps:
            program.import_(dep, True)

        proptype = f"Mapped[{idprop_pytype.mapped_pytype}]"
        propval = expr.ExprFuncCall(
            expr.ExprIdent("mapped_column"),
            [
                expr.ExprIdent(idprop_pytype.type),
                expr.ExprFuncCall(
                    expr.ExprIdent("ForeignKey"),
                    [
                        expr.ExprConstant(f"{prop.target.db.table_name}.{idprop.name}"),
                        PredefinedFn.keyword_assignment(
                            "ondelete",
                            expr.ExprConstant(prop.db.on_target_delete.to_sqlalchemy()),
                        ),
                        PredefinedFn.keyword_assignment(
                            "onupdate",
                            expr.ExprConstant(prop.db.on_target_update.to_sqlalchemy()),
                        ),
                    ],
                ),
                PredefinedFn.keyword_assignment(
                    "nullable",
                    expr.ExprConstant(prop.is_optional),
                ),
            ],
        )

        cls_ast(stmt.DefClassVarStatement(propname, proptype, propval))
        return

    # if the target class is not in the database,
    program.import_(
        f"{target_data_pkg.module(prop.target.get_pymodule_name()).path}.{prop.target.name}",
        is_import_attr=True,
    )
    propname = prop.name
    if prop.cardinality.is_star_to_many():
        if prop.is_map:
            proptype = f"Mapped[dict[str, {prop.target.name}]]"
        else:
            proptype = f"Mapped[list[{prop.target.name}]]"
    else:
        proptype = f"Mapped[{prop.target.name}]"

    # we have two choices, one is to create a composite class, one is to create a custom field
    if prop.db.is_embedded == "composite":
        # for a class to be composite, it must have only data properties
        program.import_("sqlalchemy.orm.composite", True)
        propvalargs: list[expr.Expr] = [expr.ExprIdent(prop.target.name)]
        for p in prop.target.properties.values():
            propvalargs.append(
                expr.ExprFuncCall(
                    expr.ExprIdent("mapped_column"),
                    [
                        expr.ExprConstant(f"{prop.name}_{p.name}"),
                        expr.ExprIdent(
                            assert_isinstance(p, DataProperty)
                            .datatype.get_sqlalchemy_type()
                            .type
                        ),
                        PredefinedFn.keyword_assignment(
                            "nullable",
                            expr.ExprConstant(prop.is_optional or p.is_optional),
                        ),
                    ],
                )
            )
        propval = expr.ExprFuncCall(
            expr.ExprIdent("composite"),
            propvalargs,
        )
    else:
        assert prop.db.is_embedded == "json"
        # we create a custom field, the custom field mapping need to be defined in the base
        propval = expr.ExprFuncCall(
            expr.ExprIdent("mapped_column"),
            [
                PredefinedFn.keyword_assignment(
                    "nullable",
                    expr.ExprConstant(prop.is_optional),
                ),
            ],
        )
        custom_types.append(prop)

    cls_ast(stmt.DefClassVarStatement(propname, proptype, propval))


def make_python_relational_object_property_many_to_many(
    program: Program, ast: AST, target_pkg: Package, cls: Class, prop: ObjectProperty
):
    assert cls.db is not None
    assert prop.db is not None and prop.target.db is not None
    assert prop.cardinality == Cardinality.MANY_TO_MANY

    # we create a new table to store the many-to-many relationship
    new_table = f"{cls.name}{prop.target.name}"
    clsdb = cls.db
    propdb = prop.db
    targetdb = prop.target.db

    source_idprop = assert_not_null(cls.get_id_property())
    source_id_type = source_idprop.datatype.get_python_type().type
    target_idprop = assert_not_null(prop.target.get_id_property())
    target_id_type = target_idprop.datatype.get_python_type().type

    newprogram = Program()
    newprogram.import_("__future__.annotations", True)
    newprogram.import_("sqlalchemy.ForeignKey", True)
    newprogram.import_("sqlalchemy.orm.mapped_column", True)
    newprogram.import_("sqlalchemy.orm.Mapped", True)
    newprogram.import_("sqlalchemy.orm.relationship", True)
    newprogram.import_(f"{target_pkg.path}.base.Base", True)
    newprogram.import_("typing.TYPE_CHECKING", True)
    newprogram.import_area.if_(expr.ExprIdent("TYPE_CHECKING"))(
        lambda ast00: ast00.import_(
            target_pkg.path + f".{cls.get_pymodule_name()}.{cls.name}",
            is_import_attr=True,
        ),
        lambda ast10: ast10.import_(
            target_pkg.path + f".{prop.target.get_pymodule_name()}.{prop.target.name}",
            is_import_attr=True,
        ),
    )

    newprogram.root(
        stmt.LineBreak(),
        lambda ast00: ast00.class_(new_table, [expr.ExprIdent("Base")])(
            stmt.DefClassVarStatement(
                "__tablename__",
                type=None,
                value=expr.ExprConstant(f"{clsdb.table_name}_{targetdb.table_name}"),
            ),
            stmt.LineBreak(),
            stmt.DefClassVarStatement(
                to_snake_case(cls.name),
                f"Mapped[{cls.name}]",
                expr.ExprFuncCall(
                    expr.ExprIdent("relationship"),
                    [
                        PredefinedFn.keyword_assignment(
                            "back_populates",
                            expr.ExprConstant(prop.name),
                        ),
                        PredefinedFn.keyword_assignment(
                            "lazy",
                            expr.ExprConstant("raise_on_sql"),
                        ),
                    ],
                ),
            ),
            stmt.DefClassVarStatement(
                to_snake_case(cls.name) + "_id",
                f"Mapped[{source_id_type}]",
                expr.ExprFuncCall(
                    expr.ExprIdent("mapped_column"),
                    [
                        expr.ExprFuncCall(
                            expr.ExprIdent("ForeignKey"),
                            [
                                expr.ExprConstant(
                                    f"{clsdb.table_name}.{source_idprop.name}"
                                ),
                                PredefinedFn.keyword_assignment(
                                    "ondelete",
                                    expr.ExprConstant(
                                        propdb.on_source_delete.to_sqlalchemy()
                                    ),
                                ),
                                PredefinedFn.keyword_assignment(
                                    "onupdate",
                                    expr.ExprConstant(
                                        propdb.on_source_update.to_sqlalchemy()
                                    ),
                                ),
                            ],
                        ),
                        PredefinedFn.keyword_assignment(
                            "primary_key", expr.ExprConstant(True)
                        ),
                    ],
                ),
            ),
            stmt.DefClassVarStatement(
                to_snake_case(prop.target.name),
                f"Mapped[{prop.target.name}]",
                expr.ExprFuncCall(
                    expr.ExprIdent("relationship"),
                    [
                        PredefinedFn.keyword_assignment(
                            "lazy",
                            expr.ExprConstant("raise_on_sql"),
                        ),
                    ],
                ),
            ),
            stmt.DefClassVarStatement(
                to_snake_case(prop.target.name) + "_id",
                f"Mapped[{target_id_type}]",
                expr.ExprFuncCall(
                    expr.ExprIdent("mapped_column"),
                    [
                        expr.ExprFuncCall(
                            expr.ExprIdent("ForeignKey"),
                            [
                                expr.ExprConstant(
                                    f"{targetdb.table_name}.{target_idprop.name}"
                                ),
                                PredefinedFn.keyword_assignment(
                                    "ondelete",
                                    expr.ExprConstant(
                                        propdb.on_target_delete.to_sqlalchemy()
                                    ),
                                ),
                                PredefinedFn.keyword_assignment(
                                    "onupdate",
                                    expr.ExprConstant(
                                        propdb.on_target_update.to_sqlalchemy()
                                    ),
                                ),
                            ],
                        ),
                        PredefinedFn.keyword_assignment(
                            "primary_key", expr.ExprConstant(True)
                        ),
                    ],
                ),
            ),
        ),
    )

    new_table_module = target_pkg.module(to_snake_case(new_table))
    new_table_module.write(newprogram)

    # now we add the relationship to the source.
    # we can configure it to be list, set, or dict depends on what we want.
    program.import_(new_table_module.path + f".{new_table}", True)
    program.import_("sqlalchemy.orm.relationship", True)

    # program.import_("typing.TYPE_CHECKING", True)
    # program.import_area.if_(expr.ExprIdent("TYPE_CHECKING"))(
    #     lambda ast00: ast00.import_(
    #         target_pkg.path + f".{prop.target.get_pymodule_name()}.{prop.target.name}",
    #         is_import_attr=True,
    #     )
    # )

    ast(
        stmt.DefClassVarStatement(
            prop.name,
            f"Mapped[list[{new_table}]]",
            expr.ExprFuncCall(
                expr.ExprIdent("relationship"),
                [
                    PredefinedFn.keyword_assignment(
                        "back_populates",
                        expr.ExprConstant(to_snake_case(cls.name)),
                    ),
                    PredefinedFn.keyword_assignment(
                        "lazy",
                        expr.ExprConstant("raise_on_sql"),
                    ),
                ],
            ),
        ),
    )


def get_data_conversion(
    source_pytype: str, target_pytype: str
) -> Callable[[expr.Expr], expr.Expr]:
    if source_pytype == target_pytype:
        return lambda x: x
    if source_pytype == "str" and target_pytype == "bytes":
        return lambda x: expr.ExprMethodCall(x, "encode", [])
    raise NotImplementedError(f"Cannot convert {source_pytype} to {target_pytype}")



from __future__ import annotations

import subprocess
from concurrent.futures import ThreadPoolExecutor
from dataclasses import dataclass
from pathlib import Path

from sera.typing import Language
from tqdm import tqdm


@dataclass
class File:
    path: Path
    language: Language


class Formatter:
    instance = None

    def __init__(self):
        self.pending_files: list[File] = []

    @staticmethod
    def get_instance():
        if Formatter.instance is None:
            Formatter.instance = Formatter()
        return Formatter.instance

    def register(self, file: File):
        self.pending_files.append(file)

    def process(self):
        """Format pending files in parallel"""

        def format_file(file: File):
            if file.language == Language.Typescript:
                try:
                    subprocess.check_output(
                        ["npx", "prettier", "--write", str(file.path.absolute())],
                        cwd=file.path.parent,
                    )
                except subprocess.CalledProcessError as e:
                    print(f"Error formatting {file.path}: {e}")
                    raise
            else:
                raise NotImplementedError(
                    f"Formatting not implemented for {file.language}"
                )

        if len(self.pending_files) == 0:
            return

        with ThreadPoolExecutor() as executor:
            list(
                tqdm(
                    executor.map(format_file, self.pending_files),
                    total=len(self.pending_files),
                    desc="Formatting files",
                )
            )

        self.pending_files.clear()
